from __future__ import annotations

import dataclasses
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import cached_property
from typing import Dict, List, Optional, Tuple, TypeVar, Union

from marshmallow import ValidationError
from semver import Version

from starknet_py.abi.v0 import Abi as AbiV0
from starknet_py.abi.v0 import AbiParser as AbiParserV0
from starknet_py.abi.v1 import Abi as AbiV1
from starknet_py.abi.v1 import AbiParser as AbiParserV1
from starknet_py.abi.v2 import Abi as AbiV2
from starknet_py.abi.v2 import AbiParser as AbiParserV2
from starknet_py.abi.v2.shape import (
    FUNCTION_ENTRY,
    IMPL_ENTRY,
    INTERFACE_ENTRY,
    L1_HANDLER_ENTRY,
)
from starknet_py.common import create_compiled_contract, create_sierra_compiled_contract
from starknet_py.constants import DEFAULT_DEPLOYER_ADDRESS
from starknet_py.contract_utils import _extract_compiled_class_hash, _unpack_provider
from starknet_py.hash.casm_class_hash import get_casm_hash_method_for_starknet_version
from starknet_py.hash.selector import get_selector_from_name
from starknet_py.net.account.base_account import BaseAccount
from starknet_py.net.client import Client
from starknet_py.net.client_models import (
    Call,
    EstimatedFee,
    Hash,
    ResourceBoundsMapping,
    Tag,
)
from starknet_py.net.models import AddressRepresentation, parse_address
from starknet_py.net.models.transaction import DeclareV3, InvokeV3
from starknet_py.net.udc_deployer.deployer import Deployer
from starknet_py.proxy.contract_abi_resolver import (
    ContractAbiResolver,
    ProxyConfig,
    prepare_proxy_config,
)
from starknet_py.serialization import TupleDataclass, serializer_for_function
from starknet_py.serialization.factory import serializer_for_function_v1
from starknet_py.serialization.function_serialization_adapter import (
    FunctionSerializationAdapterV0,
    FunctionSerializationAdapterV1,
)
from starknet_py.utils.constructor_args_translator import _is_abi_v2
from starknet_py.utils.sync import add_sync_methods

# pylint: disable=too-many-lines

ABI = list
ABIEntry = dict
TypeSentTransaction = TypeVar("TypeSentTransaction", bound="SentTransaction")


@dataclass(frozen=True)
class ContractData:
    """
    Basic data of a deployed contract.
    """

    address: int
    abi: ABI
    cairo_version: int

    @cached_property
    def parsed_abi(self) -> Union[AbiV0, AbiV1, AbiV2]:
        """
        Abi parsed into proper dataclass.

        :return: Abi
        """
        if self.cairo_version == 1:
            if _is_abi_v2(self.abi):
                return AbiParserV2(self.abi).parse()
            return AbiParserV1(self.abi).parse()
        return AbiParserV0(self.abi).parse()

    @staticmethod
    def from_abi(address: int, abi: ABI, cairo_version: int = 1) -> ContractData:
        """
        Create ContractData from ABI.

        :param address: Address of the deployed contract.
        :param abi: Abi of the contract.
        :param cairo_version: Version of the Cairo in which contract is written.
        :return: ContractData instance.
        """
        return ContractData(
            address=address,
            abi=abi,
            cairo_version=cairo_version,
        )


@add_sync_methods
@dataclass(frozen=True)
class SentTransaction:
    """
    Dataclass exposing the interface of transaction related to a performed action.
    """

    hash: int
    """Hash of the transaction."""

    _client: Client
    status: Optional[str] = None
    """Status of the transaction."""

    block_number: Optional[int] = None
    """Number of the block in which transaction was included."""

    async def wait_for_acceptance(
        self: TypeSentTransaction,
        check_interval: float = 2,
        retries: int = 500,
    ) -> TypeSentTransaction:
        """
        Waits for transaction to be accepted on chain till ``ACCEPTED`` status.
        Returns a new SentTransaction instance, **does not mutate original instance**.
        """

        tx_receipt = await self._client.wait_for_tx(
            self.hash,
            check_interval=check_interval,
            retries=retries,
        )
        return dataclasses.replace(
            self,
            status=tx_receipt.finality_status,
            block_number=tx_receipt.block_number,
        )


@add_sync_methods
@dataclass(frozen=True)
class InvokeResult(SentTransaction):
    """
    Result of the Invoke transaction.
    """

    # We ensure these are not None in __post_init__
    contract: ContractData = None  # pyright: ignore
    """Additional information about the Contract that made the transaction."""

    invoke_transaction: InvokeV3 = None  # pyright: ignore
    """A InvokeTransaction instance used."""

    def __post_init__(self):
        assert self.contract is not None
        assert self.invoke_transaction is not None


@add_sync_methods
@dataclass(frozen=True)
class DeclareResult(SentTransaction):
    """
    Result of the Declare transaction.
    """

    _account: BaseAccount = None  # pyright: ignore
    _cairo_version: int = 1

    class_hash: int = None  # pyright: ignore
    """Class hash of the declared contract."""

    compiled_contract: str = None  # pyright: ignore
    """Compiled contract that was declared."""

    declare_transaction: DeclareV3 = None  # pyright: ignore
    """A Declare transaction that has been sent."""

    def __post_init__(self):
        if self._account is None:
            raise ValueError("Argument _account can't be None.")

        if self.class_hash is None:
            raise ValueError("Argument class_hash can't be None.")

        if self.compiled_contract is None:
            raise ValueError("Argument compiled_contract can't be None.")

        if self.declare_transaction is None:
            raise ValueError("Argument declare_transaction can't be None.")

    async def deploy_v3(
        self,
        *,
        deployer_address: AddressRepresentation = DEFAULT_DEPLOYER_ADDRESS,
        salt: Optional[int] = None,
        unique: bool = True,
        constructor_args: Optional[Union[List, Dict]] = None,
        nonce: Optional[int] = None,
        resource_bounds: Optional[ResourceBoundsMapping] = None,
        auto_estimate: bool = False,
        tip: Optional[int] = None,
        auto_estimate_tip: bool = False,
    ) -> "DeployResult":
        """
        Deploys a contract.

        :param deployer_address: Address of the UDC. Is set to the address of
            the default UDC (same address on mainnet/sepolia) by default.
            Must be set when using custom network other than ones listed above.
        :param salt: Optional salt. Random value is selected if it is not provided.
        :param unique: Determines if the contract should be salted with the account address.
        :param constructor_args: a ``list`` or ``dict`` of arguments for the constructor.
        :param nonce: Nonce of the transaction with call to deployer.
        :param resource_bounds: Resource limits (L1 and L2) used when executing this transaction.
        :param auto_estimate: Use automatic fee estimation (not recommended, as it may lead to high costs).
        :param tip: The tip amount to be added to the transaction fee.
        :param auto_estimate_tip: Use automatic tip estimation. Using this option may lead to higher costs.
        :return: DeployResult instance.
        """
        # pylint: disable=too-many-arguments, too-many-locals
        abi = self._get_abi()

        return await Contract.deploy_contract_v3(
            account=self._account,
            class_hash=self.class_hash,
            abi=abi,
            constructor_args=constructor_args,
            deployer_address=deployer_address,
            cairo_version=self._cairo_version,
            nonce=nonce,
            resource_bounds=resource_bounds,
            auto_estimate=auto_estimate,
            salt=salt,
            unique=unique,
            tip=tip,
            auto_estimate_tip=auto_estimate_tip,
        )

    def _get_abi(self) -> List:
        if self._cairo_version == 0:
            abi = create_compiled_contract(compiled_contract=self.compiled_contract).abi
        else:
            try:
                sierra_compiled_contract = create_sierra_compiled_contract(
                    compiled_contract=self.compiled_contract
                )
                abi = sierra_compiled_contract.parsed_abi

            except Exception as exc:
                raise ValueError(
                    "Contract's ABI can't be converted to format List[Dict]. "
                    "Make sure provided compiled_contract is correct."
                ) from exc
        return abi


@add_sync_methods
@dataclass(frozen=True)
class DeployResult(SentTransaction):
    """
    Result of the contract deployment.
    """

    # We ensure this is not None in __post_init__
    deployed_contract: Contract = None  # pyright: ignore
    """A Contract instance representing the deployed contract."""

    def __post_init__(self):
        if self.deployed_contract is None:
            raise ValueError("Argument deployed_contract can't be None.")


@dataclass
class PreparedCallBase(Call):
    _client: Client
    _payload_transformer: Union[
        FunctionSerializationAdapterV0, FunctionSerializationAdapterV1
    ]


@add_sync_methods
@dataclass
class PreparedFunctionCall(PreparedCallBase):
    """
    Prepared date to call a contract function.
    """

    async def call_raw(
        self,
        block_hash: Optional[str] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> List[int]:
        """
        Calls a method without translating the result into python values.

        :param block_hash: Optional block hash.
        :param block_number: Optional block number.
        :return: list of ints.
        """
        return await self._client.call_contract(
            call=self, block_hash=block_hash, block_number=block_number
        )

    async def call(
        self,
        block_hash: Optional[str] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> Union[TupleDataclass, Tuple]:
        """
        Calls a method.

        :param block_hash: Optional block hash.
        :param block_number: Optional block number.
        :return: TupleDataclass representing call result.
        """
        result = await self.call_raw(block_hash=block_hash, block_number=block_number)
        return self._payload_transformer.deserialize(result)


@add_sync_methods
@dataclass
class PreparedFunctionInvoke(ABC, PreparedCallBase):
    _contract_data: ContractData
    _account: Optional[BaseAccount]

    def __post_init__(self):
        if self._account is None:
            raise ValueError(
                "Contract instance was created without providing an Account. "
                "It is not possible to prepare and send an invoke transaction."
            )

    @property
    def get_account(self):
        if self._account is not None:
            return self._account

        raise ValueError(
            "The account is not defined. It is not possible to send an invoke transaction."
        )

    @abstractmethod
    async def estimate_fee(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
        *,
        nonce: Optional[int] = None,
    ) -> EstimatedFee:
        """
        Estimate fee for prepared function call.

        :param block_hash: Estimate fee at specific block hash.
        :param block_number: Estimate fee at given block number
            (or "latest" / "pre_confirmed" for the latest / pre_confirmed block), default is "pre_confirmed".
        :param nonce: Nonce of the transaction.
        :return: Estimated amount of the transaction cost, either in Wei or Fri associated with executing the
            specified transaction.
        """

    async def _invoke(self, transaction: InvokeV3) -> InvokeResult:
        response = await self._client.send_transaction(transaction)

        invoke_result = InvokeResult(
            hash=response.transaction_hash,  # noinspection PyTypeChecker
            _client=self._client,
            contract=self._contract_data,
            invoke_transaction=transaction,
        )

        return invoke_result


@add_sync_methods
@dataclass
class PreparedFunctionInvokeV3(PreparedFunctionInvoke):
    """
    Prepared date to send an InvokeV3 transaction.
    """

    resource_bounds: Optional[ResourceBoundsMapping]
    tip: Optional[int] = None

    async def invoke(
        self,
        resource_bounds: Optional[ResourceBoundsMapping] = None,
        auto_estimate: bool = False,
        *,
        nonce: Optional[int] = None,
        tip: Optional[int] = None,
        auto_estimate_tip: bool = False,
    ) -> InvokeResult:
        """
        Send an Invoke transaction version 3 for the prepared data.

        :param resource_bounds: Resource limits (L1 and L2) used when executing this transaction.
        :param auto_estimate: Use automatic fee estimation (not recommended, as it may lead to high costs).
        :param nonce: Nonce of the transaction.
        :param tip: The tip amount to be added to the transaction fee.
        :param auto_estimate_tip: Use automatic tip estimation. Using this option may lead to higher costs.
        :return: InvokeResult.
        """

        transaction = await self.get_account.sign_invoke_v3(
            calls=self,
            nonce=nonce,
            resource_bounds=resource_bounds or self.resource_bounds,
            auto_estimate=auto_estimate,
            tip=tip or self.tip,
            auto_estimate_tip=auto_estimate_tip,
        )

        return await self._invoke(transaction)

    async def estimate_fee(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
        *,
        nonce: Optional[int] = None,
    ) -> EstimatedFee:
        tx = await self.get_account.sign_invoke_v3(
            calls=self,
            nonce=nonce,
            resource_bounds=ResourceBoundsMapping.init_with_zeros(),
        )
        estimate_tx = await self.get_account.sign_for_fee_estimate(transaction=tx)

        estimated_fee = await self._client.estimate_fee(
            tx=estimate_tx,
            block_hash=block_hash,
            block_number=block_number,
        )

        assert isinstance(estimated_fee, EstimatedFee)
        return estimated_fee


@add_sync_methods
class ContractFunction:
    def __init__(
        self,
        name: str,
        abi: ABIEntry,
        contract_data: ContractData,
        client: Client,
        account: Optional[BaseAccount],
        cairo_version: int = 1,
        *,
        interface_name: Optional[str] = None,
    ):
        # pylint: disable=too-many-arguments
        self.name = name
        self.abi = abi
        self.inputs = abi["inputs"]
        self.contract_data = contract_data
        self.client = client
        self.account = account

        if abi["type"] == L1_HANDLER_ENTRY:
            assert not isinstance(contract_data.parsed_abi, AbiV1)
            function = (
                contract_data.parsed_abi.l1_handler
                if contract_data.parsed_abi.l1_handler is None
                or isinstance(contract_data.parsed_abi.l1_handler, AbiV0.Function)
                else contract_data.parsed_abi.l1_handler.get(name)
            )
        elif interface_name is None:
            function = contract_data.parsed_abi.functions.get(name)
        else:
            assert isinstance(contract_data.parsed_abi, AbiV2)
            interface = contract_data.parsed_abi.interfaces[interface_name]
            function = interface.items[name]

        assert function is not None

        if cairo_version == 1:
            assert not isinstance(function, AbiV0.Function) and function is not None
            self._payload_transformer = serializer_for_function_v1(function)

        else:
            assert isinstance(function, AbiV0.Function) and function is not None
            self._payload_transformer = serializer_for_function(function)

    def prepare_call(
        self,
        *args,
        **kwargs,
    ) -> PreparedFunctionCall:
        """
        ``*args`` and ``**kwargs`` are translated into Cairo calldata.
        Creates a ``PreparedFunctionCall`` instance which exposes calldata for every argument
        and adds more arguments when calling methods.

        :return: PreparedFunctionCall.
        """

        calldata = self._payload_transformer.serialize(*args, **kwargs)
        return PreparedFunctionCall(
            to_addr=self.contract_data.address,
            calldata=calldata,
            selector=self.get_selector(self.name),
            _client=self.client,
            _payload_transformer=self._payload_transformer,
        )

    async def call(
        self,
        *args,
        block_hash: Optional[str] = None,
        block_number: Optional[Union[int, Tag]] = None,
        **kwargs,
    ) -> Union[TupleDataclass, Tuple]:
        """
        Call contract's function. ``*args`` and ``**kwargs`` are translated into Cairo calldata.
        The result is translated from Cairo data to python values.
        Equivalent of ``.prepare_call(*args, **kwargs).call()``.

        :param block_hash: Block hash to perform the call to the contract at specific point of time.
        :param block_number: Block number to perform the call to the contract at specific point of time.
        :return: TupleDataclass representing call result.
        """
        return await self.prepare_call(*args, **kwargs).call(
            block_hash=block_hash, block_number=block_number
        )

    def prepare_invoke_v3(
        self,
        *args,
        resource_bounds: Optional[ResourceBoundsMapping] = None,
        tip: Optional[int] = None,
        **kwargs,
    ) -> PreparedFunctionInvokeV3:
        """
        ``*args`` and ``**kwargs`` are translated into Cairo calldata.
        Creates a ``PreparedFunctionInvokeV3`` instance which exposes calldata for every argument
        and adds more arguments when calling methods.

        :param resource_bounds: Resource limits (L1 and L2) used when executing this transaction.
        :param tip: The tip amount to be added to the transaction fee.
        :return: PreparedFunctionInvokeV3.
        """

        calldata = self._payload_transformer.serialize(*args, **kwargs)
        return PreparedFunctionInvokeV3(
            to_addr=self.contract_data.address,
            calldata=calldata,
            selector=self.get_selector(self.name),
            resource_bounds=resource_bounds,
            tip=tip,
            _contract_data=self.contract_data,
            _client=self.client,
            _account=self.account,
            _payload_transformer=self._payload_transformer,
        )

    async def invoke_v3(
        self,
        *args,
        resource_bounds: Optional[ResourceBoundsMapping] = None,
        auto_estimate: bool = False,
        tip: Optional[int] = None,
        auto_estimate_tip: bool = False,
        nonce: Optional[int] = None,
        **kwargs,
    ) -> InvokeResult:
        """
        Invoke contract's function. ``*args`` and ``**kwargs`` are translated into Cairo calldata.
        Equivalent of ``.prepare_invoke_v3(*args, **kwargs).invoke()``.

        :param resource_bounds: Resource limits (L1 and L2) used when executing this transaction.
        :param auto_estimate: Use automatic fee estimation (not recommended, as it may lead to high costs).
        :param tip: The tip amount to be added to the transaction fee.
        :param auto_estimate_tip: Use automatic tip estimation. Using this option may lead to higher costs.
        :param nonce: Nonce of the transaction.
        :return: InvokeResult.
        """
        prepared_invoke = self.prepare_invoke_v3(*args, **kwargs)
        return await prepared_invoke.invoke(
            resource_bounds=resource_bounds,
            nonce=nonce,
            auto_estimate=auto_estimate,
            tip=tip,
            auto_estimate_tip=auto_estimate_tip,
        )

    @staticmethod
    def get_selector(function_name: str):
        """
        :param function_name: Contract function's name.
        :return: A Starknet integer selector for this function inside the contract.
        """
        return get_selector_from_name(function_name)


FunctionsRepository = Dict[str, ContractFunction]


@add_sync_methods
class Contract:
    """
    Cairo contract's model.
    """

    def __init__(
        self,
        address: AddressRepresentation,
        abi: list,
        provider: Union[BaseAccount, Client],
        *,
        cairo_version: int = 1,
    ):
        """
        Should be used instead of ``from_address`` when ABI is known statically.

        Arguments provider and client are mutually exclusive and cannot be provided at the same time.

        :param address: contract's address.
        :param abi: contract's abi.
        :param provider: BaseAccount or Client used to perform transactions.
        :param cairo_version: Version of the Cairo in which contract is written.
        """
        client, account = _unpack_provider(provider)

        self.account: Optional[BaseAccount] = account
        self.client: Client = client
        self.data = ContractData.from_abi(parse_address(address), abi, cairo_version)

        try:
            self._functions = self._make_functions(
                contract_data=self.data,
                client=self.client,
                account=self.account,
                cairo_version=cairo_version,
            )
        except ValidationError as exc:
            raise ValueError(
                "Make sure valid ABI is used to create a Contract instance"
            ) from exc

    @property
    def functions(self) -> FunctionsRepository:
        """
        :return: All functions exposed from a contract.
        """
        return self._functions

    @property
    def address(self) -> int:
        """Address of the contract."""
        return self.data.address

    @staticmethod
    async def from_address(
        address: AddressRepresentation,
        provider: Union[BaseAccount, Client] = None,  # pyright: ignore
        proxy_config: Union[bool, ProxyConfig] = False,
    ) -> Contract:
        """
        Fetches ABI for given contract and creates a new Contract instance with it. If you know ABI statically you
        should create Contract's instances directly instead of using this function to avoid unnecessary API calls.

        :raises ContractNotFoundError: when contract is not found.
        :raises TypeError: when given client's `get_class_by_hash` method does not return abi.
        :raises ProxyResolutionError: when given ProxyChecks were not sufficient to resolve proxy's implementation.
        :param address: Contract's address.
        :param provider: BaseAccount or Client.
        :param proxy_config: Proxy resolving config
            If set to ``True``, will use default proxy checks
            :class:`starknet_py.proxy.proxy_check.OpenZeppelinProxyCheck` and
            :class:`starknet_py.proxy.proxy_check.ArgentProxyCheck`.

            If set to ``False``, :meth:`Contract.from_address` will not resolve proxies.

            If a valid :class:`starknet_py.contract_abi_resolver.ProxyConfig` is provided, will use its values instead.

        :return: an initialized Contract instance.
        """
        client, account = _unpack_provider(provider)

        address = parse_address(address)
        proxy_config = Contract._create_proxy_config(proxy_config)

        abi, cairo_version = await ContractAbiResolver(
            address=address, client=client, proxy_config=proxy_config
        ).resolve()

        return Contract(
            address=address,
            abi=abi,
            provider=account or client,
            cairo_version=cairo_version,
        )

    @staticmethod
    async def declare_v3(
        account: BaseAccount,
        compiled_contract: str,
        *,
        compiled_contract_casm: Optional[str] = None,
        compiled_class_hash: Optional[int] = None,
        nonce: Optional[int] = None,
        resource_bounds: Optional[ResourceBoundsMapping] = None,
        auto_estimate: bool = False,
        tip: Optional[int] = None,
        auto_estimate_tip: bool = False,
    ) -> DeclareResult:
        # pylint: disable=too-many-arguments

        """
        Declares a contract.

        :param account: BaseAccount used to sign and send declare transaction.
        :param compiled_contract: String containing compiled contract.
        :param compiled_contract_casm: String containing the content of the starknet-sierra-compile (.casm file).
        :param compiled_class_hash: Hash of the compiled_contract_casm.
        :param nonce: Nonce of the transaction.
        :param resource_bounds: Resource limits (L1 and L2) used when executing this transaction.
        :param auto_estimate: Use automatic fee estimation (not recommended, as it may lead to high costs).
        :param tip: The tip amount to be added to the transaction fee.
        :param auto_estimate_tip: Use automatic tip estimation. Using this option may lead to higher costs.
        :return: DeclareResult instance.
        """

        block = await account.client.get_block()
        starknet_version = Version.parse(block.starknet_version)
        hash_method = get_casm_hash_method_for_starknet_version(starknet_version)

        compiled_class_hash = _extract_compiled_class_hash(
            compiled_contract_casm, compiled_class_hash, hash_method=hash_method
        )

        declare_tx = await account.sign_declare_v3(
            compiled_contract=compiled_contract,
            compiled_class_hash=compiled_class_hash,
            nonce=nonce,
            resource_bounds=resource_bounds,
            auto_estimate=auto_estimate,
            tip=tip,
            auto_estimate_tip=auto_estimate_tip,
        )

        return await _declare_contract(
            declare_tx, account, compiled_contract, cairo_version=1
        )

    @staticmethod
    async def deploy_contract_v3(
        account: BaseAccount,
        class_hash: Hash,
        abi: Optional[List] = None,
        constructor_args: Optional[Union[List, Dict]] = None,
        *,
        deployer_address: AddressRepresentation = DEFAULT_DEPLOYER_ADDRESS,
        cairo_version: int = 1,
        nonce: Optional[int] = None,
        resource_bounds: Optional[ResourceBoundsMapping] = None,
        auto_estimate: bool = False,
        salt: Optional[int] = None,
        unique: bool = True,
        tip: Optional[int] = None,
        auto_estimate_tip: bool = False,
    ) -> "DeployResult":
        """
        Deploys a contract through Universal Deployer Contract.

        :param account: BaseAccount used to sign and send deploy transaction.
        :param class_hash: The class_hash of the contract to be deployed.
        :param abi: An abi of the contract to be deployed.
        :param constructor_args: a ``list`` or ``dict`` of arguments for the constructor.
        :param deployer_address: Address of the UDC. Is set to the address of
            the default UDC (same address on mainnet/sepolia) by default.
            Must be set when using custom network other than ones listed above.
        :param cairo_version: Version of the Cairo in which contract is written.
            By default, it is set to 1.
        :param nonce: Nonce of the transaction.
        :param resource_bounds: Resource limits (L1 and L2) used when executing this transaction.
        :param auto_estimate: Use automatic fee estimation (not recommended, as it may lead to high costs).
        :param salt: Optional salt. Random value is selected if it is not provided.
        :param unique: Determines if the contract should be salted with the account address.
        :param tip: The tip amount to be added to the transaction fee.
        :param auto_estimate_tip: Use automatic tip estimation. Using this option may lead to higher costs.
        :return: DeployResult instance.
        """
        # pylint: disable=too-many-arguments, too-many-locals
        deployer = Deployer(
            deployer_address=deployer_address,
            account_address=account.address if unique else None,
        )
        deploy_call, address = deployer.create_contract_deployment(
            class_hash=class_hash,
            salt=salt,
            abi=abi,
            calldata=constructor_args,
            cairo_version=cairo_version,
        )

        res = await account.execute_v3(
            calls=deploy_call,
            nonce=nonce,
            resource_bounds=resource_bounds,
            auto_estimate=auto_estimate,
            tip=tip,
            auto_estimate_tip=auto_estimate_tip,
        )

        if abi is None:
            contract_class = await account.client.get_class_by_hash(class_hash)
            abi = ContractAbiResolver.get_abi_from_contract_class(contract_class)

        deployed_contract = Contract(
            provider=account, address=address, abi=abi, cairo_version=cairo_version
        )

        deploy_result = DeployResult(
            hash=res.transaction_hash,
            _client=account.client,
            deployed_contract=deployed_contract,
        )

        return deploy_result

    @classmethod
    def _make_functions(
        cls,
        contract_data: ContractData,
        client: Client,
        account: Optional[BaseAccount],
        cairo_version: int = 1,
    ) -> FunctionsRepository:
        repository = {}
        implemented_interfaces = [
            entry["interface_name"]
            for entry in contract_data.abi
            if entry["type"] == IMPL_ENTRY
        ]

        for abi_entry in contract_data.abi:
            if abi_entry["type"] in [FUNCTION_ENTRY, L1_HANDLER_ENTRY]:
                name = abi_entry["name"]
                repository[name] = ContractFunction(
                    name=name,
                    abi=abi_entry,
                    contract_data=contract_data,
                    client=client,
                    account=account,
                    cairo_version=cairo_version,
                )

            if (
                abi_entry["type"] == INTERFACE_ENTRY
                and abi_entry["name"] in implemented_interfaces
            ):
                for item in abi_entry["items"]:
                    name = item["name"]
                    repository[name] = ContractFunction(
                        name=name,
                        abi=item,
                        contract_data=contract_data,
                        client=client,
                        account=account,
                        cairo_version=cairo_version,
                        interface_name=abi_entry["name"],
                    )

        return repository

    @staticmethod
    def _create_proxy_config(proxy_config) -> ProxyConfig:
        if proxy_config is False:
            return ProxyConfig()
        proxy_arg = ProxyConfig() if proxy_config is True else proxy_config
        return prepare_proxy_config(proxy_arg)


async def _declare_contract(
    transaction: DeclareV3,
    account: BaseAccount,
    compiled_contract: str,
    cairo_version: int,
) -> DeclareResult:
    res = await account.client.declare(transaction=transaction)

    return DeclareResult(
        hash=res.transaction_hash,
        class_hash=res.class_hash,
        compiled_contract=compiled_contract,
        declare_transaction=transaction,
        _account=account,
        _client=account.client,
        _cairo_version=cairo_version,
    )



from collections import OrderedDict

import pytest

from starknet_py.serialization.data_serializers.felt_serializer import FeltSerializer
from starknet_py.serialization.data_serializers.payload_serializer import (
    PayloadSerializer,
)
from starknet_py.serialization.errors import InvalidTypeException, InvalidValueException
from starknet_py.serialization.function_serialization_adapter import (
    FunctionSerializationAdapterV0,
)
from starknet_py.serialization.tuple_dataclass import TupleDataclass

serializer = FunctionSerializationAdapterV0(
    inputs_serializer=PayloadSerializer(
        OrderedDict(
            x=FeltSerializer(),
            y=FeltSerializer(),
            z=FeltSerializer(),
        )
    ),
    outputs_deserializer=PayloadSerializer(
        OrderedDict(
            a=FeltSerializer(),
            b=FeltSerializer(),
            c=FeltSerializer(),
        )
    ),
)


def test_serialize():
    serialized = [1, 2, 3]
    # pylint: disable=invalid-name
    x, y, z = serialized

    assert serializer.serialize(x, y, z) == serialized
    assert serializer.serialize(x=x, y=y, z=z) == serialized
    assert serializer.serialize(x, y=y, z=z) == serialized

    with pytest.raises(
        InvalidTypeException, match="Provided 4 positional arguments, 3 possible."
    ):
        serializer.serialize(1, 2, 3, 4)

    with pytest.raises(
        InvalidTypeException,
        match="Both positional and named argument provided for 'x'.",
    ):
        serializer.serialize(1, 2, x=1, y=3)

    with pytest.raises(
        InvalidTypeException,
        match="Unnecessary named arguments provided: 'unknown_key'.",
    ):
        serializer.serialize(1, 2, unknown_key=3)

    with pytest.raises(InvalidTypeException, match="Missing arguments: 'z'."):
        serializer.serialize(1, 2)


def test_deserialize():
    assert serializer.deserialize([1, 2, 3]) == TupleDataclass.from_dict(
        {"a": 1, "b": 2, "c": 3}
    )

    with pytest.raises(
        InvalidValueException,
        match="Last 2 values '0x4,0x5' out of total 5 values were not used during deserialization.",
    ):
        serializer.deserialize([1, 2, 3, 4, 5])

    with pytest.raises(
        InvalidValueException,
        match="Last 4 values '0x4,0x5,0x6...' out of total 7 values were not used during deserialization.",
    ):
        serializer.deserialize([1, 2, 3, 4, 5, 6, 7])

    with pytest.raises(
        InvalidValueException,
        match="Not enough data to deserialize 'c'. Can't read 1 values at position 2, 0 available.",
    ):
        serializer.deserialize([1, 2])



from dataclasses import dataclass, field
from typing import Generator, List, Tuple

from starknet_py.serialization._context import (
    DeserializationContext,
    SerializationContext,
)
from starknet_py.serialization.data_serializers.cairo_data_serializer import (
    CairoDataSerializer,
)


@dataclass
class OutputSerializer(CairoDataSerializer[List, Tuple]):
    """
    Serializer for function output.
    Can't serialize anything.
    Deserializes data to a Tuple.

    Example:
        [1, 1, 1] => (340282366920938463463374607431768211457)
    """

    serializers: List[CairoDataSerializer] = field(init=True)

    def deserialize_with_context(self, context: DeserializationContext) -> Tuple:
        result = []

        for index, serializer in enumerate(self.serializers):
            with context.push_entity("output[" + str(index) + "]"):
                result.append(serializer.deserialize_with_context(context))

        return tuple(result)

    def serialize_with_context(
        self, context: SerializationContext, value: List
    ) -> Generator[int, None, None]:
        raise ValueError(
            "Output serializer can't be used to transform python data into calldata."
        )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Set, Tuple, cast

from starknet_py.cairo.felt import CairoData
from starknet_py.serialization.data_serializers.output_serializer import (
    OutputSerializer,
)
from starknet_py.serialization.data_serializers.payload_serializer import (
    PayloadSerializer,
)
from starknet_py.serialization.errors import InvalidTypeException
from starknet_py.serialization.tuple_dataclass import TupleDataclass


@dataclass
class FunctionSerializationAdapter:
    """
    Class serializing ``*args`` and ``**kwargs`` by adapting them to function inputs.
    """

    inputs_serializer: PayloadSerializer

    expected_args: Tuple[str] = field(init=False)

    def __post_init__(self):
        self.expected_args = tuple(
            self.inputs_serializer.serializers.keys()
        )  # pyright: ignore

    def serialize(self, *args, **kwargs) -> CairoData:
        """
        Method using args and kwargs to match members and serialize them separately.

        :return: Members serialized separately in SerializedPayload.
        """
        named_arguments = self._merge_arguments(args, kwargs)
        return self.inputs_serializer.serialize(named_arguments)

    def _merge_arguments(self, args: Tuple, kwargs: Dict) -> Dict:
        """
        Merges positional and keyed arguments.
        """
        # After this line we know that len(args) <= len(self.expected_args)
        self._ensure_no_unnecessary_positional_args(args)

        named_arguments = dict(kwargs)
        for arg, input_name in zip(args, self.expected_args):
            if input_name in kwargs:
                raise InvalidTypeException(
                    f"Both positional and named argument provided for '{input_name}'."
                )
            named_arguments[input_name] = arg

        expected_args = set(self.expected_args)
        provided_args = set(named_arguments.keys())

        # named_arguments might have unnecessary arguments coming from kwargs (we ensure that
        # len(args) <= len(self.expected_args) above)
        self._ensure_no_unnecessary_args(expected_args, provided_args)

        # there might be some argument missing (not provided)
        self._ensure_no_missing_args(expected_args, provided_args)

        return named_arguments

    def _ensure_no_unnecessary_positional_args(self, args: Tuple):
        if len(args) > len(self.expected_args):
            raise InvalidTypeException(
                f"Provided {len(args)} positional arguments, {len(self.expected_args)} possible."
            )

    @staticmethod
    def _ensure_no_unnecessary_args(expected_args: Set[str], provided_args: Set[str]):
        excessive_arguments = provided_args - expected_args
        if excessive_arguments:
            raise InvalidTypeException(
                f"Unnecessary named arguments provided: '{', '.join(excessive_arguments)}'."
            )

    @staticmethod
    def _ensure_no_missing_args(expected_args: Set[str], provided_args: Set[str]):
        missing_arguments = expected_args - provided_args
        if missing_arguments:
            raise InvalidTypeException(
                f"Missing arguments: '{', '.join(missing_arguments)}'."
            )


@dataclass
class FunctionSerializationAdapterV0(FunctionSerializationAdapter):
    """
    Class serializing ``*args`` and ``**kwargs`` by adapting them to function inputs.
    """

    outputs_deserializer: PayloadSerializer

    def deserialize(self, data: List[int]) -> TupleDataclass:
        """
        Deserializes data into TupleDataclass containing python representations.

        :return: cairo data.
        """
        return self.outputs_deserializer.deserialize(data)


@dataclass
class FunctionSerializationAdapterV1(FunctionSerializationAdapter):
    outputs_deserializer: OutputSerializer

    def deserialize(self, data: List[int]) -> Tuple:
        """
        Deserializes data into TupleDataclass containing python representations.

        :return: cairo data.
        """
        return cast(Tuple, self.outputs_deserializer.deserialize(data))



from typing import cast

import pytest

from starknet_py.constants import FIELD_PRIME
from starknet_py.serialization.data_serializers.felt_serializer import FeltSerializer
from starknet_py.serialization.errors import InvalidTypeException, InvalidValueException


@pytest.mark.parametrize(
    "value",
    [0, 1, 322132123, FIELD_PRIME - 1],
)
def test_valid_felt_values(value):
    serialized = FeltSerializer().serialize(value)
    deserialized = FeltSerializer().deserialize([value])

    assert deserialized == value
    assert serialized == [value]


def test_valid_shortstring():
    with pytest.warns(DeprecationWarning, match="Serializing shortstrings"):
        serialized = FeltSerializer().serialize(
            # We use cast because we don't want to show that strings are accepted in types
            cast(int, "string shorter than 32 chars")
        )
        assert serialized == [
            0x737472696E672073686F72746572207468616E203332206368617273
        ]


def test_invalid_shortstrings():
    with pytest.raises(InvalidValueException, match="Expected an ascii string"):
        FeltSerializer().serialize(cast(int, "õ"))

    with pytest.raises(
        InvalidValueException, match="cannot be longer than 31 characters"
    ):
        FeltSerializer().serialize(cast(int, "a" * 32))


def test_invalid_type():
    error_message = "Error: expected int, received '{}' of type '<class 'dict'>."
    with pytest.raises(InvalidTypeException, match=error_message):
        FeltSerializer().serialize(cast(int, {}))


@pytest.mark.parametrize(
    "value",
    [-1, -100, FIELD_PRIME, FIELD_PRIME + 10000],
)
def test_values_out_of_range(value):
    error_message = f"Error: invalid value '{value}' - must be in .* range."
    with pytest.raises(InvalidValueException, match=error_message):
        FeltSerializer().serialize(value)
    with pytest.raises(InvalidValueException, match=error_message):
        FeltSerializer().deserialize([value])


def test_not_all_values_used():
    error_message = "Last 1 values '0x2' out of total 2 values were not used during deserialization."
    with pytest.raises(InvalidValueException, match=error_message):
        FeltSerializer().deserialize([1, 2])



import dataclasses
from collections import OrderedDict
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from starknet_py.common import create_sierra_compiled_contract
from starknet_py.constants import (
    ANY_CALLER,
    QUERY_VERSION_BASE,
    STRK_FEE_CONTRACT_ADDRESS,
    OutsideExecutionInterfaceID,
)
from starknet_py.hash.address import compute_address
from starknet_py.hash.outside_execution import outside_execution_to_typed_data
from starknet_py.hash.selector import get_selector_from_name
from starknet_py.hash.utils import verify_message_signature
from starknet_py.net.account.account_deployment_result import AccountDeploymentResult
from starknet_py.net.account.base_account import (
    BaseAccount,
    OutsideExecutionSupportBaseMixin,
)
from starknet_py.net.client import Client
from starknet_py.net.client_models import (
    Call,
    Calls,
    EstimatedFee,
    Hash,
    OutsideExecution,
    OutsideExecutionTimeBounds,
    ResourceBoundsMapping,
    SentTransactionResponse,
    SierraContractClass,
    Tag,
)
from starknet_py.net.full_node_client import FullNodeClient
from starknet_py.net.models import AddressRepresentation, parse_address
from starknet_py.net.models.chains import RECOGNIZED_CHAIN_IDS, Chain, parse_chain
from starknet_py.net.models.transaction import (
    AccountTransaction,
    DeclareV3,
    DeployAccountV3,
    InvokeV3,
    TypeAccountTransaction,
)
from starknet_py.net.models.typed_data import TypedDataDict
from starknet_py.net.signer import BaseSigner
from starknet_py.net.signer.key_pair import KeyPair
from starknet_py.net.signer.stark_curve_signer import StarkCurveSigner
from starknet_py.net.tip import estimate_tip
from starknet_py.serialization.data_serializers import (
    ArraySerializer,
    FeltSerializer,
    PayloadSerializer,
    StructSerializer,
    UintSerializer,
)
from starknet_py.utils.iterable import ensure_iterable
from starknet_py.utils.sync import add_sync_methods
from starknet_py.utils.typed_data import TypedData


# pylint: disable=too-many-public-methods,disable=too-many-lines
@add_sync_methods
class Account(BaseAccount, OutsideExecutionSupportBaseMixin):
    """
    Default Account implementation.
    """

    ESTIMATED_FEE_MULTIPLIER: float = 1.5
    """Amount by which each estimated fee is multiplied when using `auto_estimate`."""

    ESTIMATED_AMOUNT_MULTIPLIER: float = 1.5
    ESTIMATED_UNIT_PRICE_MULTIPLIER: float = 1.5
    """Values by which each estimated `max_amount` and `max_price_per_unit` are multiplied when using 
    `auto_estimate`. Used only for V3 transactions"""

    def __init__(
        self,
        *,
        address: AddressRepresentation,
        client: Client,
        signer: Optional[BaseSigner] = None,
        key_pair: Optional[KeyPair] = None,
        chain: Optional[Chain] = None,
    ):
        """
        :param address: Address of the account contract.
        :param client: Instance of Client which will be used to add transactions.
        :param signer: Custom signer to be used by Account.
                       If none is provided, default
                       :py:class:`starknet_py.net.signer.stark_curve_signer.StarkCurveSigner` is used.
        :param key_pair: Key pair that will be used to create a default `Signer`.
        :param chain: Chain ID associated with the account.
            This can be supplied in multiple formats:

            - an enum :py:class:`starknet_py.net.models.StarknetChainId`
            - a string name (e.g. 'SN_SEPOLIA')
            - a hexadecimal value (e.g. '0x1')
            - an integer (e.g. 1)
        """
        self._address = parse_address(address)
        self._client = client
        self._cairo_version = None
        self._chain_id = None if chain is None else parse_chain(chain)

        if signer is not None and key_pair is not None:
            raise ValueError("Arguments signer and key_pair are mutually exclusive.")

        if signer is None:
            if key_pair is None:
                raise ValueError(
                    "Either a signer or a key_pair must be provided in Account constructor."
                )
            if self._chain_id is None:
                raise ValueError("One of chain or signer must be provided.")

            signer = StarkCurveSigner(
                account_address=self.address, key_pair=key_pair, chain_id=self._chain_id
            )
        self.signer: BaseSigner = signer

    @property
    def address(self) -> int:
        return self._address

    @property
    async def cairo_version(self) -> int:
        if self._cairo_version is None:
            assert isinstance(self._client, FullNodeClient)
            contract_class = await self._client.get_class_at(
                contract_address=self._address
            )
            self._cairo_version = (
                1 if isinstance(contract_class, SierraContractClass) else 0
            )
        return self._cairo_version

    @property
    def client(self) -> Client:
        return self._client

    async def _get_tip(self, *, tip: Optional[int], auto_estimate_tip: bool) -> int:
        if auto_estimate_tip is True and tip is not None:
            raise ValueError(
                "Arguments tip and auto_estimate_tip are mutually exclusive."
            )

        if auto_estimate_tip:
            return await estimate_tip(self.client)

        if tip is None:
            return 0

        return tip

    async def _get_resource_bounds(
        self,
        transaction: AccountTransaction,
        resource_bounds: Optional[ResourceBoundsMapping] = None,
        auto_estimate: bool = False,
    ) -> ResourceBoundsMapping:
        if auto_estimate and resource_bounds is not None:
            raise ValueError(
                "Arguments auto_estimate and resource_bounds are mutually exclusive."
            )

        if auto_estimate:
            estimated_fee = await self.estimate_fee(transaction)
            assert isinstance(estimated_fee, EstimatedFee)

            return estimated_fee.to_resource_bounds()

        if resource_bounds is None:
            raise ValueError(
                "One of arguments: resource_bounds or auto_estimate must be specified when invoking a transaction."
            )

        return resource_bounds

    async def _prepare_invoke_v3(
        self,
        calls: Calls,
        *,
        resource_bounds: Optional[ResourceBoundsMapping] = None,
        nonce: Optional[int] = None,
        auto_estimate: bool = False,
        tip: Optional[int] = None,
        auto_estimate_tip: bool = False,
    ) -> InvokeV3:
        # pylint: disable=too-many-arguments
        """
        Takes calls and creates InvokeV3 from them.

        :param calls: Single call or a list of calls.
        :param resource_bounds: Resource limits (L1 and L2) that can be used in this transaction.
        :param auto_estimate: Use automatic fee estimation; not recommended as it may lead to high costs.
        :return: InvokeV3 created from the calls (without the signature).
        """
        if nonce is None:
            nonce = await self.get_nonce()

        wrapped_calldata = _parse_calls(await self.cairo_version, calls)

        tip = await self._get_tip(tip=tip, auto_estimate_tip=auto_estimate_tip)

        transaction = InvokeV3(
            calldata=wrapped_calldata,
            resource_bounds=ResourceBoundsMapping.init_with_zeros(),
            signature=[],
            nonce=nonce,
            sender_address=self.address,
            version=3,
            tip=tip,
        )

        resource_bounds = await self._get_resource_bounds(
            transaction, resource_bounds, auto_estimate
        )
        return _add_resource_bounds_to_transaction(transaction, resource_bounds)

    async def estimate_fee(
        self,
        tx: Union[AccountTransaction, List[AccountTransaction]],
        skip_validate: bool = False,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> Union[EstimatedFee, List[EstimatedFee]]:
        transactions = (
            await self.sign_for_fee_estimate(tx)
            if isinstance(tx, AccountTransaction)
            else [await self.sign_for_fee_estimate(t) for t in tx]
        )

        return await self._client.estimate_fee(
            tx=transactions,
            skip_validate=skip_validate,
            block_hash=block_hash,
            block_number=block_number,
        )

    async def get_nonce(
        self,
        *,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> int:
        """
        Get the current nonce of the account.

        :param block_hash: Block's hash or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :param block_number: Block's number or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :return: nonce.
        """
        return await self._client.get_contract_nonce(
            self.address, block_hash=block_hash, block_number=block_number
        )

    async def _check_outside_execution_nonce(
        self,
        nonce: int,
        *,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> bool:
        (is_valid,) = await self._client.call_contract(
            call=Call(
                to_addr=self.address,
                selector=get_selector_from_name("is_valid_outside_execution_nonce"),
                calldata=[nonce],
            ),
            block_hash=block_hash,
            block_number=block_number,
        )
        return bool(is_valid)

    async def get_outside_execution_nonce(self, retry_count=10) -> int:
        while retry_count > 0:
            random_stark_address = KeyPair.generate().public_key
            if await self._check_outside_execution_nonce(random_stark_address):
                return random_stark_address
            retry_count -= 1
        raise RuntimeError("Failed to generate a valid nonce")

    async def _get_outside_execution_version(
        self,
    ) -> Union[OutsideExecutionInterfaceID, None]:
        for version in [
            OutsideExecutionInterfaceID.V1,
            OutsideExecutionInterfaceID.V2,
        ]:
            if await self.supports_interface(version):
                return version
        return None

    async def supports_interface(
        self, interface_id: OutsideExecutionInterfaceID
    ) -> bool:
        (does_support,) = await self._client.call_contract(
            Call(
                to_addr=self.address,
                selector=get_selector_from_name("supports_interface"),
                calldata=[interface_id],
            )
        )
        return bool(does_support)

    async def get_balance(
        self,
        token_address: Optional[AddressRepresentation] = None,
        *,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> int:
        if token_address is None:
            chain_id = await self._get_chain_id()
            if chain_id in RECOGNIZED_CHAIN_IDS:
                token_address = STRK_FEE_CONTRACT_ADDRESS
            else:
                raise ValueError(
                    "Argument token_address must be specified when using a custom network."
                )

        low, high = await self._client.call_contract(
            Call(
                to_addr=parse_address(token_address),
                selector=get_selector_from_name("balance_of"),
                calldata=[self.address],
            ),
            block_hash=block_hash,
            block_number=block_number,
        )

        return (high << 128) + low

    async def sign_for_fee_estimate(
        self, transaction: TypeAccountTransaction
    ) -> TypeAccountTransaction:
        version = transaction.version + QUERY_VERSION_BASE
        transaction = dataclasses.replace(transaction, version=version)

        signature = self.signer.sign_transaction(transaction)
        return _add_signature_to_transaction(tx=transaction, signature=signature)

    async def sign_outside_execution_call(
        self,
        calls: Calls,
        execution_time_bounds: OutsideExecutionTimeBounds,
        *,
        caller: AddressRepresentation = ANY_CALLER,
        nonce: Optional[int] = None,
        interface_version: Optional[OutsideExecutionInterfaceID] = None,
    ) -> Call:
        if interface_version is None:
            interface_version = await self._get_outside_execution_version()

        if interface_version is None:
            raise RuntimeError(
                "Can't initiate call, outside execution is not supported."
            )

        if nonce is None:
            nonce = await self.get_outside_execution_nonce()

        outside_execution = OutsideExecution(
            caller=parse_address(caller),
            nonce=nonce,
            execute_after=execution_time_bounds.execute_after_timestamp,
            execute_before=execution_time_bounds.execute_before_timestamp,
            calls=list(ensure_iterable(calls)),
        )
        chain_id = await self._get_chain_id()
        signature = self.signer.sign_message(
            outside_execution_to_typed_data(
                outside_execution, interface_version, chain_id
            ),
            self.address,
        )
        selector_for_version = {
            OutsideExecutionInterfaceID.V1: "execute_from_outside",
            OutsideExecutionInterfaceID.V2: "execute_from_outside_v2",
        }

        return Call(
            to_addr=self.address,
            selector=get_selector_from_name(selector_for_version[interface_version]),
            calldata=_outside_transaction_serialiser.serialize(
                {
                    "outside_execution": outside_execution.to_abi_dict(),
                    "signature": signature,
                }
            ),
        )

    async def sign_invoke_v3(
        self,
        calls: Calls,
        *,
        nonce: Optional[int] = None,
        resource_bounds: Optional[ResourceBoundsMapping] = None,
        auto_estimate: bool = False,
        tip: Optional[int] = None,
        auto_estimate_tip: bool = False,
    ) -> InvokeV3:
        # pylint: disable=too-many-arguments
        invoke_tx = await self._prepare_invoke_v3(
            calls,
            resource_bounds=resource_bounds,
            nonce=nonce,
            auto_estimate=auto_estimate,
            tip=tip,
            auto_estimate_tip=auto_estimate_tip,
        )
        signature = self.signer.sign_transaction(invoke_tx)
        return _add_signature_to_transaction(invoke_tx, signature)

    async def sign_declare_v3(
        self,
        compiled_contract: str,
        compiled_class_hash: int,
        *,
        nonce: Optional[int] = None,
        resource_bounds: Optional[ResourceBoundsMapping] = None,
        auto_estimate: bool = False,
        tip: Optional[int] = None,
        auto_estimate_tip: bool = False,
    ) -> DeclareV3:
        # pylint: disable=too-many-arguments
        declare_tx = await self._make_declare_v3_transaction(
            compiled_contract,
            compiled_class_hash,
            nonce=nonce,
            tip=tip,
            auto_estimate_tip=auto_estimate_tip,
        )
        resource_bounds = await self._get_resource_bounds(
            declare_tx, resource_bounds, auto_estimate
        )
        declare_tx = _add_resource_bounds_to_transaction(declare_tx, resource_bounds)

        signature = self.signer.sign_transaction(declare_tx)
        return _add_signature_to_transaction(declare_tx, signature)

    async def _make_declare_v3_transaction(
        self,
        compiled_contract: str,
        compiled_class_hash: int,
        *,
        nonce: Optional[int] = None,
        tip: Optional[int] = None,
        auto_estimate_tip: bool = False,
    ) -> DeclareV3:
        contract_class = create_sierra_compiled_contract(
            compiled_contract=compiled_contract
        )

        if nonce is None:
            nonce = await self.get_nonce()

        tip = await self._get_tip(tip=tip, auto_estimate_tip=auto_estimate_tip)

        declare_tx = DeclareV3(
            contract_class=contract_class.convert_to_sierra_contract_class(),
            compiled_class_hash=compiled_class_hash,
            sender_address=self.address,
            signature=[],
            nonce=nonce,
            version=3,
            resource_bounds=ResourceBoundsMapping.init_with_zeros(),
            tip=tip,
        )
        return declare_tx

    async def sign_deploy_account_v3(
        self,
        class_hash: int,
        contract_address_salt: int,
        *,
        constructor_calldata: Optional[List[int]] = None,
        nonce: int = 0,
        resource_bounds: Optional[ResourceBoundsMapping] = None,
        auto_estimate: bool = False,
        tip: Optional[int] = None,
        auto_estimate_tip: bool = False,
    ) -> DeployAccountV3:
        # pylint: disable=too-many-arguments
        tip = await self._get_tip(tip=tip, auto_estimate_tip=auto_estimate_tip)

        deploy_account_tx = DeployAccountV3(
            class_hash=class_hash,
            contract_address_salt=contract_address_salt,
            constructor_calldata=(constructor_calldata or []),
            version=3,
            resource_bounds=ResourceBoundsMapping.init_with_zeros(),
            signature=[],
            nonce=nonce,
            tip=tip,
        )
        resource_bounds = await self._get_resource_bounds(
            deploy_account_tx, resource_bounds, auto_estimate
        )
        deploy_account_tx = _add_resource_bounds_to_transaction(
            deploy_account_tx, resource_bounds
        )

        signature = self.signer.sign_transaction(deploy_account_tx)
        return _add_signature_to_transaction(deploy_account_tx, signature)

    async def execute_v3(
        self,
        calls: Calls,
        *,
        resource_bounds: Optional[ResourceBoundsMapping] = None,
        nonce: Optional[int] = None,
        auto_estimate: bool = False,
        tip: Optional[int] = None,
        auto_estimate_tip: bool = False,
    ) -> SentTransactionResponse:
        # pylint: disable=too-many-arguments
        execute_transaction = await self.sign_invoke_v3(
            calls,
            nonce=nonce,
            resource_bounds=resource_bounds,
            auto_estimate=auto_estimate,
            tip=tip,
            auto_estimate_tip=auto_estimate_tip,
        )
        return await self._client.send_transaction(execute_transaction)

    def sign_message(self, typed_data: Union[TypedData, TypedDataDict]) -> List[int]:
        if isinstance(typed_data, TypedData):
            return self.signer.sign_message(typed_data, self.address)
        typed_data_dataclass = TypedData.from_dict(typed_data)
        return self.signer.sign_message(typed_data_dataclass, self.address)

    def verify_message(
        self, typed_data: Union[TypedData, TypedDataDict], signature: List[int]
    ) -> bool:
        if not isinstance(typed_data, TypedData):
            typed_data = TypedData.from_dict(typed_data)
        message_hash = typed_data.message_hash(account_address=self.address)
        return verify_message_signature(message_hash, signature, self.signer.public_key)

    @staticmethod
    async def deploy_account_v3(
        *,
        address: AddressRepresentation,
        class_hash: int,
        salt: int,
        key_pair: KeyPair,
        client: Client,
        constructor_calldata: Optional[List[int]] = None,
        nonce: int = 0,
        resource_bounds: Optional[ResourceBoundsMapping] = None,
        auto_estimate: bool = False,
        tip: Optional[int] = None,
        auto_estimate_tip: bool = False,
    ) -> AccountDeploymentResult:
        # pylint: disable=too-many-arguments, too-many-locals

        """
        Deploys an account contract with provided class_hash on Starknet and returns
        an AccountDeploymentResult that allows waiting for transaction acceptance.

        Provided address must be first prefunded with enough tokens, otherwise the method will fail.

        :param address: Calculated and prefunded address of the new account.
        :param class_hash: Class hash of the account contract to be deployed.
        :param salt: Salt used to calculate the address.
        :param key_pair: KeyPair used to calculate address and sign deploy account transaction.
        :param client: Client instance used for deployment.
        :param constructor_calldata: Optional calldata to account contract constructor. If ``None`` is passed,
            ``[key_pair.public_key]`` will be used as calldata.
        :param nonce: Nonce of the transaction.
        :param resource_bounds: Resource limits (L1 and L2) used when executing this transaction.
        :param auto_estimate: Use automatic fee estimation, not recommend as it may lead to high costs.
        :param auto_estimate_tip: Use automatic tip estimation. Using this option may lead to higher costs.
        :param tip: The tip amount to be added to the transaction fee.
        """
        calldata = (
            constructor_calldata
            if constructor_calldata is not None
            else [key_pair.public_key]
        )

        chain = await client.get_chain_id()

        account = _prepare_account_to_deploy(
            address=address,
            class_hash=class_hash,
            salt=salt,
            key_pair=key_pair,
            client=client,
            chain=chain,
            calldata=calldata,
        )

        deploy_account_tx = await account.sign_deploy_account_v3(
            class_hash=class_hash,
            contract_address_salt=salt,
            constructor_calldata=calldata,
            nonce=nonce,
            resource_bounds=resource_bounds,
            auto_estimate=auto_estimate,
            tip=tip,
            auto_estimate_tip=auto_estimate_tip,
        )

        result = await client.deploy_account(deploy_account_tx)

        return AccountDeploymentResult(
            hash=result.transaction_hash, account=account, _client=account.client
        )

    async def _get_chain_id(self) -> int:
        if self._chain_id is None:
            chain = await self._client.get_chain_id()
            self._chain_id = parse_chain(chain)

        return self._chain_id


def _prepare_account_to_deploy(
    address: AddressRepresentation,
    class_hash: int,
    salt: int,
    key_pair: KeyPair,
    client: Client,
    chain: Chain,
    calldata: List[int],
) -> Account:
    # pylint: disable=too-many-arguments
    address = parse_address(address)

    if address != (
        computed := compute_address(
            salt=salt,
            class_hash=class_hash,
            constructor_calldata=calldata,
            deployer_address=0,
        )
    ):
        raise ValueError(
            f"Provided address {hex(address)} is different than computed address {hex(computed)} "
            f"for the given class_hash and salt."
        )

    return Account(
        address=address,
        client=client,
        key_pair=key_pair,
        chain=chain,
    )


def _is_sierra_contract(data: Dict[str, Any]) -> bool:
    return "sierra_program" in data


def _add_signature_to_transaction(
    tx: TypeAccountTransaction, signature: List[int]
) -> TypeAccountTransaction:
    return dataclasses.replace(tx, signature=signature)


def _add_max_fee_to_transaction(
    tx: TypeAccountTransaction, max_fee: int
) -> TypeAccountTransaction:
    return dataclasses.replace(tx, max_fee=max_fee)


def _add_resource_bounds_to_transaction(
    tx: TypeAccountTransaction, resource_bounds: ResourceBoundsMapping
) -> TypeAccountTransaction:
    return dataclasses.replace(tx, resource_bounds=resource_bounds)


def _parse_calls(cairo_version: int, calls: Calls) -> List[int]:
    if cairo_version == 1:
        parsed_calls = _parse_calls_cairo_v1(ensure_iterable(calls))
        wrapped_calldata = _execute_payload_serializer_v1.serialize(
            {"calls": parsed_calls}
        )
    else:
        call_descriptions, calldata = _merge_calls(ensure_iterable(calls))
        wrapped_calldata = _execute_payload_serializer_v0.serialize(
            {"call_array": call_descriptions, "calldata": calldata}
        )
    return wrapped_calldata


def _parse_call_cairo_v0(call: Call, entire_calldata: List) -> Tuple[Dict, List]:
    _data = {
        "to": call.to_addr,
        "selector": call.selector,
        "data_offset": len(entire_calldata),
        "data_len": len(call.calldata),
    }
    entire_calldata += call.calldata

    return _data, entire_calldata


def _merge_calls(calls: Iterable[Call]) -> Tuple[List[Dict], List[int]]:
    call_descriptions = []
    entire_calldata = []
    for call in calls:
        data, entire_calldata = _parse_call_cairo_v0(call, entire_calldata)
        call_descriptions.append(data)

    return call_descriptions, entire_calldata


def _parse_calls_cairo_v1(calls: Iterable[Call]) -> List[Dict]:
    calls_parsed = []
    for call in calls:
        _data = {
            "to": call.to_addr,
            "selector": call.selector,
            "calldata": call.calldata,
        }
        calls_parsed.append(_data)

    return calls_parsed


_felt_serializer = FeltSerializer()
_call_description_cairo_v0 = StructSerializer(
    OrderedDict(
        to=_felt_serializer,
        selector=_felt_serializer,
        data_offset=_felt_serializer,
        data_len=_felt_serializer,
    )
)
_call_description_cairo_v1 = StructSerializer(
    OrderedDict(
        to=_felt_serializer,
        selector=_felt_serializer,
        calldata=ArraySerializer(_felt_serializer),
    )
)

_execute_payload_serializer_v0 = PayloadSerializer(
    OrderedDict(
        call_array=ArraySerializer(_call_description_cairo_v0),
        calldata=ArraySerializer(_felt_serializer),
    )
)
_execute_payload_serializer_v1 = PayloadSerializer(
    OrderedDict(
        calls=ArraySerializer(_call_description_cairo_v1),
    )
)
_outside_transaction_serialiser = StructSerializer(
    OrderedDict(
        outside_execution=StructSerializer(
            OrderedDict(
                caller=FeltSerializer(),
                nonce=FeltSerializer(),
                execute_after=UintSerializer(bits=64),
                execute_before=UintSerializer(bits=64),
                calls=ArraySerializer(_call_description_cairo_v1),
            )
        ),
        signature=ArraySerializer(FeltSerializer()),
    )
)



import re
from abc import ABC
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Optional, Union, cast

from marshmallow import Schema, fields, post_load

from starknet_py.cairo.felt import encode_shortstring
from starknet_py.constants import FIELD_PRIME
from starknet_py.hash.hash_method import HashMethod
from starknet_py.hash.selector import get_selector_from_name
from starknet_py.net.client_utils import _to_rpc_felt
from starknet_py.net.models.typed_data import DomainDict, Revision, TypedDataDict
from starknet_py.net.schemas.common import RevisionField
from starknet_py.serialization.data_serializers import ByteArraySerializer
from starknet_py.utils.merkle_tree import MerkleTree


@dataclass(frozen=True)
class Parameter(ABC):
    name: str
    type: str


@dataclass(frozen=True)
class StandardParameter(Parameter):
    pass


@dataclass(frozen=True)
class MerkleTreeParameter(Parameter):
    type: str = field(default="merkletree", init=False)
    contains: str


@dataclass(frozen=True)
class EnumParameter(Parameter):
    type: str = field(default="enum", init=False)
    contains: str


@dataclass
class Domain:
    """
    Dataclass representing a domain object (StarkNetDomain, StarknetDomain)
    """

    name: str
    version: str
    chain_id: Union[str, int]
    revision: Optional[Revision] = None

    def __post_init__(self):
        self.resolved_revision = (
            Revision(self.revision) if self.revision else Revision.V0
        )
        self.separator_name = self._resolve_separator_name()

    def _resolve_separator_name(self):
        if self.resolved_revision == Revision.V0:
            return "StarkNetDomain"
        return "StarknetDomain"

    @staticmethod
    def from_dict(data: DomainDict) -> "Domain":
        """
        Create Domain dataclass from dictionary.

        :param data: Domain dictionary.
        :return: Domain dataclass instance.
        """
        return cast(Domain, DomainSchema().load(data))

    def to_dict(self) -> dict:
        """
        Create Domain dictionary from dataclass.

        :return: Domain dictionary.
        """
        return cast(Dict, DomainSchema().dump(obj=self))


@dataclass(frozen=True)
class TypeContext:
    """
    Dataclass representing a Context object
    """

    parent: str
    key: str


class BasicType(Enum):
    FELT = "felt"
    SELECTOR = "selector"
    MERKLE_TREE = "merkletree"
    ENUM = "enum"
    SHORT_STRING = "shortstring"
    STRING = "string"
    CONTRACT_ADDRESS = "ContractAddress"
    CLASS_HASH = "ClassHash"
    BOOL = "bool"
    U128 = "u128"
    I128 = "i128"
    TIMESTAMP = "timestamp"


class PresetType(Enum):
    U256 = "u256"
    TOKEN_AMOUNT = "TokenAmount"
    NFT_ID = "NftId"


@dataclass(frozen=True)
class TypedData:
    """
    Dataclass representing a TypedData object
    """

    types: Dict[str, List[Parameter]]
    primary_type: str
    domain: Domain
    message: dict

    def __post_init__(self):
        self._verify_types()

    @property
    def _all_types(self):
        preset_types = _get_preset_types(self.domain.resolved_revision)
        return {
            **preset_types,
            **self.types,
        }

    @property
    def _hash_method(self) -> HashMethod:
        if self.domain.resolved_revision == Revision.V0:
            return HashMethod.PEDERSEN
        return HashMethod.POSEIDON

    @staticmethod
    def from_dict(data: TypedDataDict) -> "TypedData":
        """
        Create TypedData dataclass from dictionary.

        :param data: TypedData dictionary.
        :return: TypedData dataclass instance.
        """
        return cast(TypedData, TypedDataSchema().load(data))

    def to_dict(self) -> dict:
        """
        Create TypedData dictionary from dataclass.

        :return: TypedData dictionary.
        """

        return cast(Dict, TypedDataSchema().dump(obj=self))

    def _is_struct(self, type_name: str) -> bool:
        return type_name in self.types

    def _encode_value_v1(
        self,
        basic_type: BasicType,
        value: Union[int, str, dict],
        type_name: str,
        context: Optional[TypeContext] = None,
    ) -> Optional[int]:
        if basic_type in (
            BasicType.FELT,
            BasicType.SHORT_STRING,
            BasicType.CONTRACT_ADDRESS,
            BasicType.CLASS_HASH,
        ) and isinstance(value, (int, str)):
            return parse_felt(value)

        if basic_type in (
            BasicType.U128,
            BasicType.TIMESTAMP,
        ) and isinstance(value, (int, str)):
            return encode_u128(value)

        if basic_type == BasicType.I128 and isinstance(value, (int, str)):
            return encode_i128(value)

        if basic_type == BasicType.STRING and isinstance(value, str):
            return self._encode_long_string(value)

        if basic_type == BasicType.ENUM and isinstance(value, dict):
            if context is None:
                raise ValueError(f"Context is not provided for '{type_name}' type.")
            return self._encode_enum(value, context)

        return None

    # pylint: disable=no-self-use
    def _encode_value_v0(
        self,
        basic_type: BasicType,
        value: Union[int, str],
    ) -> Optional[int]:
        if basic_type in (
            BasicType.FELT,
            BasicType.STRING,
        ) and isinstance(value, (int, str)):
            return parse_felt(value)

        return None

    def _encode_value(
        self,
        type_name: str,
        value: Union[int, str, dict, list],
        context: Optional[TypeContext] = None,
    ) -> int:
        if type_name in self._all_types and isinstance(value, dict):
            return self.struct_hash(type_name, value)

        if is_pointer(type_name) and isinstance(value, list):
            type_name = strip_pointer(type_name)
            hashes = [self._encode_value(type_name, val) for val in value]
            return self._hash_method.hash_many(hashes)

        if type_name not in _get_basic_type_names(self.domain.resolved_revision):
            raise ValueError(f"Type [{type_name}] is not defined in types.")

        basic_type = BasicType(type_name)

        encoded_value = None
        if self.domain.resolved_revision == Revision.V0 and isinstance(
            value, (str, int)
        ):
            encoded_value = self._encode_value_v0(basic_type, value)
        elif self.domain.resolved_revision == Revision.V1 and isinstance(
            value, (str, int, dict)
        ):
            encoded_value = self._encode_value_v1(basic_type, value, type_name, context)

        if encoded_value is not None:
            return encoded_value

        if basic_type == BasicType.BOOL and isinstance(value, (bool, str, int)):
            return encode_bool(value)

        if basic_type == BasicType.SELECTOR and isinstance(value, str):
            return prepare_selector(value)

        if basic_type == BasicType.MERKLE_TREE and isinstance(value, list):
            if context is None:
                raise ValueError(f"Context is not provided for '{type_name}' type.")
            return self._prepare_merkle_tree_root(value, context)

        raise ValueError(
            f"Error occurred while encoding value with type name {type_name}."
        )

    def _encode_data(self, type_name: str, data: dict) -> List[int]:
        values = []
        for param in self._all_types[type_name]:
            encoded_value = self._encode_value(
                param.type,
                data[param.name],
                TypeContext(parent=type_name, key=param.name),
            )
            values.append(encoded_value)

        return values

    # pylint: disable=too-many-branches
    def _verify_types(self):
        if self.domain.separator_name not in self.types:
            raise ValueError(f"Types must contain '{self.domain.separator_name}'.")

        referenced_types = set()
        for type_name in self.types:
            for ref_type in self.types[type_name]:
                if isinstance(ref_type, (EnumParameter, MerkleTreeParameter)):
                    referenced_types.add(ref_type.contains)
                elif is_enum_variant_type(ref_type.type):
                    referenced_types.update(_extract_enum_types(ref_type.type))
                else:
                    referenced_types.add(strip_pointer(ref_type.type))

        referenced_types.update([self.domain.separator_name, self.primary_type])

        basic_type_names = _get_basic_type_names(self.domain.resolved_revision)
        preset_type_names = _get_preset_types(self.domain.resolved_revision).keys()

        for type_name in self.types:
            if not type_name:
                raise ValueError("Type names cannot be empty.")

            if type_name in basic_type_names:
                raise ValueError(
                    f"Types must not contain basic types. [{type_name}] was found."
                )

            if type_name in preset_type_names:
                raise ValueError(
                    f"Types must not contain preset types. [{type_name}] was found."
                )

            if is_pointer(type_name):
                raise ValueError(
                    f"Type names cannot end in *. [{type_name}] was found."
                )

            if is_enum_variant_type(type_name):
                raise ValueError(
                    f"Type names cannot be enclosed in parentheses. [{type_name}] was found."
                )

            if "," in type_name:
                raise ValueError(
                    f"Type names cannot contain commas. [{type_name}] was found."
                )

            if type_name not in referenced_types:
                raise ValueError(
                    f"Dangling types are not allowed. Unreferenced type [{type_name}] was found."
                )

            for ref_type in self.types[type_name]:
                if isinstance(ref_type, EnumParameter):
                    self._validate_enum_type()

    def _validate_enum_type(self):
        if self.domain.resolved_revision != Revision.V1:
            raise ValueError(
                f"'{BasicType.ENUM.name}' basic type is not supported in revision "
                f"{self.domain.resolved_revision.value}."
            )

    def _get_dependencies(self, type_name: str) -> List[str]:
        dependencies = [type_name]
        to_visit = [type_name]

        while to_visit:
            current_type = to_visit.pop(0)
            params = self._all_types.get(current_type, [])

            for param in params:
                if isinstance(param, EnumParameter):
                    extracted_types = [param.contains]
                elif is_enum_variant_type(param.type):
                    extracted_types = _extract_enum_types(param.type)
                else:
                    extracted_types = [param.type]

                extracted_types = [
                    strip_pointer(extr_type) for extr_type in extracted_types
                ]
                for extracted_type in extracted_types:
                    if (
                        extracted_type in self._all_types
                        and extracted_type not in dependencies
                    ):
                        dependencies.append(extracted_type)
                        to_visit.append(extracted_type)

        return list(dependencies)

    def _encode_type(self, type_name: str) -> str:
        primary, *dependencies = self._get_dependencies(type_name)
        types = [primary, *sorted(dependencies)]

        def encode_dependency(dependency: str) -> str:
            def escape(s: str) -> str:
                if self.domain.resolved_revision == Revision.V0:
                    return s
                return f'"{s}"'

            if dependency not in self._all_types:
                raise ValueError(f"Dependency [{dependency}] is not defined in types.")

            encoded_params = []
            for param in self._all_types[dependency]:
                target_type = (
                    param.contains
                    if isinstance(param, EnumParameter)
                    and self.domain.resolved_revision == Revision.V1
                    else param.type
                )

                if is_enum_variant_type(target_type):
                    type_str = _extract_enum_types(target_type)
                    type_str = f"({','.join([escape(x) for x in type_str])})"

                else:
                    type_str = escape(target_type)

                encoded_params.append(f"{escape(param.name)}:{type_str}")
            encoded_params = ",".join(encoded_params)

            return f"{escape(dependency)}({encoded_params})"

        return "".join([encode_dependency(x) for x in types])

    def type_hash(self, type_name: str) -> int:
        """
        Calculate the hash of a type name.

        :param type_name: Name of the type.
        :return: Hash of the type name.
        """
        return get_selector_from_name(self._encode_type(type_name))

    def struct_hash(self, type_name: str, data: dict) -> int:
        """
        Calculate the hash of a struct.

        :param type_name: Name of the type.
        :param data: Data defining the struct.
        :return: Hash of the struct.
        """
        return self._hash_method.hash_many(
            [self.type_hash(type_name), *self._encode_data(type_name, data)]
        )

    def message_hash(self, account_address: int) -> int:
        """
        Calculate the hash of the message.

        :param account_address: Address of an account.
        :return: Hash of the message.
        """
        message = [
            encode_shortstring("StarkNet Message"),
            self.struct_hash(self.domain.separator_name, self.domain.to_dict()),
            account_address,
            self.struct_hash(self.primary_type, self.message),
        ]

        return self._hash_method.hash_many(message)

    def _prepare_merkle_tree_root(self, value: List, context: TypeContext) -> int:
        merkle_tree_type = self._get_merkle_tree_leaves_type(context)
        struct_hashes = [
            self._encode_value(merkle_tree_type, struct) for struct in value
        ]

        return MerkleTree(struct_hashes, self._hash_method).root_hash

    def _get_merkle_tree_leaves_type(self, context: TypeContext) -> str:
        target_type = self._resolve_type(context)

        if not isinstance(target_type, MerkleTreeParameter):
            raise ValueError("Target type is not a merkletree type.")

        return target_type.contains

    def _resolve_type(self, context: TypeContext) -> Parameter:
        parent, key = context.parent, context.key

        if parent not in self._all_types:
            raise ValueError(f"Parent {parent} is not defined in types.")

        parent_type = self._all_types[parent]

        target_type = next((item for item in parent_type if item.name == key), None)
        if target_type is None:
            raise ValueError(
                f"Key {key} is not defined in type {parent} or multiple definitions are present."
            )

        if not isinstance(target_type, (EnumParameter, MerkleTreeParameter)):
            raise ValueError("Target type is not an enum or merkletree type.")
        return target_type

    def _encode_enum(self, value: dict, context: TypeContext):
        if len(value.keys()) != 1:
            raise ValueError(
                f"'{BasicType.ENUM.name}' value must contain a single variant."
            )

        variant_name, variant_data = next(iter(value.items()))
        variants = self._get_enum_variants(context)
        variant_definition = next(
            (item for item in variants if item.name == variant_name), None
        )

        if variant_definition is None:
            raise ValueError(
                f"Variant [{variant_name}] is not defined in '${BasicType.ENUM.name}' "
                f"type [{context.key}] or multiple definitions are present."
            )

        encoded_subtypes = []
        extracted_enum_types = _extract_enum_types(variant_definition.type)

        for i, subtype in enumerate(extracted_enum_types):
            subtype_data = variant_data[i]
            encoded_subtypes.append(self._encode_value(subtype, subtype_data))

        variant_index = variants.index(variant_definition)
        return self._hash_method.hash_many([variant_index, *encoded_subtypes])

    def _get_enum_variants(self, context: TypeContext) -> List[Parameter]:
        enum_type = self._resolve_type(context)
        if not isinstance(enum_type, EnumParameter):
            raise ValueError(f"Type [{context.key}] is not an enum.")
        if enum_type.contains not in self._all_types:
            raise ValueError(f"Type [{enum_type.contains}] is not defined in types")

        return self._all_types[enum_type.contains]

    def _encode_long_string(self, value: str) -> int:
        byte_array_serializer = ByteArraySerializer()
        serialized_values = byte_array_serializer.serialize(value)
        return self._hash_method.hash_many(serialized_values)


def _extract_enum_types(value: str) -> List[str]:
    if not is_enum_variant_type(value):
        raise ValueError(f"Type [{value}] is not an enum.")

    value = value[1:-1]
    if not value:
        return []

    return value.split(",")


def parse_felt(value: Union[int, str]) -> int:
    if isinstance(value, int):
        return value
    if value.startswith("0x"):
        return int(value, 16)
    if value.isnumeric():
        return int(value)
    return encode_shortstring(value)


def is_pointer(value: str) -> bool:
    return value.endswith("*")


def strip_pointer(value: str) -> str:
    if is_pointer(value):
        return value[:-1]
    return value


def is_enum_variant_type(value: str) -> bool:
    return value.startswith("(") and value.endswith(")")


def prepare_selector(name: str) -> int:
    try:
        return int(_to_rpc_felt(name), 16)
    except ValueError:
        return get_selector_from_name(name)


def encode_bool(value: Union[bool, str, int]) -> int:
    if isinstance(value, bool):
        return 1 if value else 0
    if isinstance(value, int) and value in (0, 1):
        return value
    if isinstance(value, str) and value in ("0", "1"):
        return int(value)
    if isinstance(value, str) and value in ("false", "true"):
        return 0 if value == "false" else 1
    if isinstance(value, str) and value in ("0x0", "0x1"):
        return int(value, 16)
    raise ValueError(f"Expected boolean value, got [{value}].")


def is_digit_string(s: str, signed=False) -> bool:
    if signed:
        return bool(re.fullmatch(r"-?\d+", s))
    return bool(re.fullmatch(r"\d+", s))


def encode_u128(value: Union[str, int]) -> int:
    def is_in_range(n: int):
        return 0 <= n < 2**128

    if isinstance(value, str) and value.startswith("0x"):
        int_value = int(value, 16)
    elif isinstance(value, str) and is_digit_string(value):
        int_value = int(value)
    elif isinstance(value, int):
        int_value = value
    else:
        raise ValueError(f"Value [{value}] is not a valid number.")

    if is_in_range(int_value):
        return int_value
    raise ValueError(f"Value [{value}] is out of range for '{BasicType.U128}'.")


def encode_i128(value: Union[str, int]) -> int:
    def is_in_range(n: int):
        return (n < 2**127) or (n >= (FIELD_PRIME - (2**127)))

    if isinstance(value, str) and value.startswith("0x"):
        int_value = int(value, 16)
    elif isinstance(value, str) and is_digit_string(value, True):
        int_value = int(value)
    elif isinstance(value, int):
        int_value = value
    else:
        raise ValueError(f"Value [{value}] is not a valid number.")

    if abs(int_value) >= FIELD_PRIME:
        raise ValueError(
            f"Values outside the range (-FIELD_PRIME, FIELD_PRIME) are not allowed, [{value}] given."
        )
    int_value %= FIELD_PRIME

    if is_in_range(int_value):
        return int_value
    raise ValueError(f"Value [{value}] is out of range for '{BasicType.I128}'.")


def _get_basic_type_names(revision: Revision) -> List[str]:
    basic_types_v0 = [
        BasicType.FELT,
        BasicType.SELECTOR,
        BasicType.MERKLE_TREE,
        BasicType.STRING,
        BasicType.BOOL,
    ]

    basic_types_v1 = list(BasicType)

    basic_types = basic_types_v0 if revision == Revision.V0 else basic_types_v1
    return [basic_type.value for basic_type in basic_types]


def _get_preset_types(
    revision: Revision,
) -> Dict[str, List[StandardParameter]]:
    if revision == Revision.V0:
        return {}

    return {
        PresetType.U256.value: [
            StandardParameter(name="low", type="u128"),
            StandardParameter(name="high", type="u128"),
        ],
        PresetType.TOKEN_AMOUNT.value: [
            StandardParameter(name="token_address", type="ContractAddress"),
            StandardParameter(name="amount", type="u256"),
        ],
        PresetType.NFT_ID.value: [
            StandardParameter(name="collection_address", type="ContractAddress"),
            StandardParameter(name="token_id", type="u256"),
        ],
    }


# pylint: disable=unused-argument
# pylint: disable=no-self-use


class ParameterSchema(Schema):
    name = fields.String(data_key="name", required=True)
    type = fields.String(data_key="type", required=True)
    contains = fields.String(data_key="contains", required=False)

    @post_load
    def make_dataclass(self, data, **kwargs) -> Parameter:
        type_val = data["type"]

        if type_val == BasicType.MERKLE_TREE.value:
            return MerkleTreeParameter(name=data["name"], contains=data["contains"])

        if type_val == BasicType.ENUM.value:
            return EnumParameter(name=data["name"], contains=data["contains"])

        return StandardParameter(name=data["name"], type=type_val)


class DomainSchema(Schema):
    name = fields.String(data_key="name", required=True)
    version = fields.String(data_key="version", required=True)
    chain_id = fields.String(attribute="chain_id", data_key="chainId", required=True)
    revision = RevisionField(data_key="revision", required=False)

    @post_load
    def make_dataclass(self, data, **kwargs) -> Domain:
        return Domain(
            name=data["name"],
            version=data["version"],
            chain_id=data["chain_id"],
            revision=data.get("revision"),
        )


class TypedDataSchema(Schema):
    types = fields.Dict(
        data_key="types",
        keys=fields.Str(),
        values=fields.List(fields.Nested(ParameterSchema())),
    )
    primary_type = fields.String(data_key="primaryType", required=True)
    domain = fields.Nested(DomainSchema, required=True)
    message = fields.Dict(data_key="message", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> TypedData:
        return TypedData(
            types=data["types"],
            primary_type=data["primary_type"],
            domain=data["domain"],
            message=data["message"],
        )



import pytest


@pytest.mark.asyncio
async def test_executing_transactions(account, map_contract):
    address = map_contract.address
    # pylint: disable=import-outside-toplevel
    # docs: start
    from starknet_py.hash.selector import get_selector_from_name
    from starknet_py.net.client_models import (
        Call,
        ResourceBounds,
        ResourceBoundsMapping,
    )

    call = Call(
        to_addr=address, selector=get_selector_from_name("put"), calldata=[20, 20]
    )

    resp = await account.execute_v3(
        calls=call,
        resource_bounds=ResourceBoundsMapping(
            l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
            l2_gas=ResourceBounds(max_amount=int(1e9), max_price_per_unit=int(1e17)),
            l1_data_gas=ResourceBounds(
                max_amount=int(1e5), max_price_per_unit=int(1e13)
            ),
        ),
    )

    await account.client.wait_for_tx(resp.transaction_hash)
    # docs: end

    call = Call(to_addr=address, selector=get_selector_from_name("get"), calldata=[20])
    (value,) = await account.client.call_contract(call)
    assert value == 20



# data from cairo repository: crates/cairo-lang-starknet/src/abi_test.rs
from collections import OrderedDict

from starknet_py.abi.v1.model import Abi
from starknet_py.cairo.data_types import (
    ArrayType,
    EnumType,
    FeltType,
    StructType,
    UintType,
)

core_structures = {
    "core::starknet::eth_address::EthAddress": StructType(
        name="core::starknet::eth_address::EthAddress",
        types=OrderedDict([("address", FeltType())]),
    )
}

pool_id_dict = {
    "type": "struct",
    "name": "PoolId",
    "members": [
        {"name": "value", "type": "core::integer::u256"},
    ],
}
pool_id_struct = StructType("PoolId", OrderedDict(value=UintType(256)))

user_dict = {
    "type": "struct",
    "name": "User",
    "members": [
        {"name": "id", "type": "core::integer::u256"},
        {"name": "name_len", "type": "core::felt252"},
        {"name": "name", "type": "core::array::Array::<core::felt252>"},
        {"name": "pool_id", "type": "PoolId"},
    ],
}
user_struct = StructType(
    "User",
    OrderedDict(
        id=UintType(256),
        name_len=FeltType(),
        name=ArrayType(FeltType()),
        pool_id=pool_id_struct,
    ),
)

user_added_dict = {
    "type": "event",
    "name": "UserAdded",
    "inputs": [
        {"name": "user", "type": "User"},
    ],
}
user_added_event: Abi.Event = Abi.Event(
    "UserAdded",
    OrderedDict(user=user_struct),
)

pool_id_added_dict = {
    "type": "event",
    "name": "PoolIdAdded",
    "inputs": [
        {"name": "pool_id", "type": "PoolId"},
    ],
}
pool_id_added_event: Abi.Event = Abi.Event(
    "PoolIdAdded",
    OrderedDict(pool_id=pool_id_struct),
)

get_user_dict = {
    "type": "function",
    "name": "get_user",
    "inputs": [
        {
            "name": "id",
            "type": "core::integer::u256",
        }
    ],
    "outputs": [{"type": "User"}],
}
get_user_fn = Abi.Function("get_user", OrderedDict(id=UintType(256)), [user_struct])

delete_pool_dict = {
    "type": "function",
    "name": "delete_pool",
    "inputs": [
        {
            "name": "id",
            "type": "PoolId",
        },
        {"name": "user_id", "type": "core::integer::u256"},
    ],
    "outputs": [],
}
delete_pool_fn = Abi.Function(
    "delete_pool", OrderedDict(id=pool_id_struct, user_id=UintType(256)), []
)

my_struct_dict = {
    "type": "struct",
    "name": "test::MyStruct::<core::integer::u256>",
    "members": [
        {"name": "a", "type": "core::integer::u256"},
        {"name": "b", "type": "core::felt252"},
    ],
}
my_struct = StructType(
    name="test::MyStruct::<core::integer::u256>",
    types=OrderedDict(a=UintType(256), b=FeltType()),
)

my_enum_dict = {
    "type": "enum",
    "name": "test::MyEnum::<core::integer::u128>",
    "variants": [
        {"name": "a", "type": "core::integer::u256"},
        {"name": "b", "type": "test::MyStruct::<core::integer::u128>"},
    ],
}
my_enum = EnumType(
    name="test::MyEnum::<core::integer::u128>",
    variants=OrderedDict(a=UintType(256), b=my_struct),
)

foo_event_dict = {
    "type": "event",
    "name": "foo_event",
    "inputs": [
        {"name": "a", "type": "core::felt252"},
        {"name": "b", "type": "core::integer::u128"},
    ],
}
foo_event = Abi.Event(
    name="foo_event", inputs=OrderedDict(a=FeltType(), b=UintType(128))
)

foo_external_dict = {
    "type": "function",
    "name": "foo_external",
    "inputs": [
        {"name": "a", "type": "core::felt252"},
        {"name": "b", "type": "core::integer::u128"},
    ],
    "outputs": [{"type": "test::MyStruct::<core::integer::u256>"}],
    "state_mutability": "external",
}
foo_external = Abi.Function(
    name="foo_external",
    inputs=OrderedDict(a=FeltType(), b=UintType(128)),
    outputs=[my_struct],
)

foo_view_dict = {
    "type": "function",
    "name": "foo_view",
    "inputs": [
        {"name": "a", "type": "core::felt252"},
        {"name": "b", "type": "core::integer::u128"},
    ],
    "outputs": [{"type": "test::MyEnum::<core::integer::u128>"}],
    "state_mutability": "view",
}
foo_view = Abi.Function(
    name="foo_view",
    inputs=OrderedDict(a=FeltType(), b=UintType(128)),
    outputs=[my_enum],
)

abi_v1 = Abi(
    defined_structures={},
    events={
        "PoolIdAdded": pool_id_added_event,
    },
    functions={},
    defined_enums={},
)

event_abi_v1 = [
    {
        "type": "event",
        "name": "Approval",
        "inputs": [
            {
                "name": "owner",
                "type": "core::starknet::contract_address::ContractAddress",
            },
            {
                "name": "spender",
                "type": "core::starknet::contract_address::ContractAddress",
            },
            {"name": "value", "type": "core::integer::u256"},
        ],
    }
]



from __future__ import annotations

from typing import Dict, Union

from starknet_py.abi.v1.parser_transformer import parse
from starknet_py.cairo.data_types import CairoType, EnumType, StructType, TypeIdentifier


class UnknownCairoTypeError(ValueError):
    """
    Error thrown when TypeParser finds type that was not declared prior to parsing.
    """

    type_name: str

    def __init__(self, type_name: str):
        super().__init__(
            # pylint: disable=line-too-long
            f"Type '{type_name}' is not defined. Please report this issue at https://github.com/software-mansion/starknet.py/issues"
        )
        self.type_name = type_name


class TypeParser:
    """
    Low level utility class for parsing Cairo types that can be used in external methods.
    """

    defined_types: Dict[str, Union[StructType, EnumType]]

    def __init__(self, defined_types: Dict[str, Union[StructType, EnumType]]):
        """
        TypeParser constructor.

        :param defined_types: dictionary containing all defined types. For now, they can only be structures.
        """
        self.defined_types = defined_types
        for name, defined_type in defined_types.items():
            if name != defined_type.name:
                raise ValueError(
                    f"Keys must match name of type, '{name}' != '{defined_type.name}'."
                )

    def parse_inline_type(self, type_string: str) -> CairoType:
        """
        Inline type is one that can be used inline, for instance as return type. For instance
        (core::felt252, (), (core::felt252,)). Structure can only be referenced in inline type, can't be defined
        this way.

        :param type_string: type to parse.
        """
        parsed = parse(type_string, self.defined_types)
        if isinstance(parsed, TypeIdentifier):
            for defined_name in self.defined_types.keys():
                if parsed.name == defined_name.split("<")[0].strip(":"):
                    return self.defined_types[defined_name]
            raise UnknownCairoTypeError(parsed.name)

        return parsed



"""
TypedDict structures for TypedData
"""

import sys
from typing import Any, Dict, List, Optional, TypedDict

from starknet_py.net.schemas.common import Revision

if sys.version_info < (3, 11):
    from typing_extensions import NotRequired
else:
    from typing import NotRequired


class ParameterDict(TypedDict):
    """
    TypedDict representing a Parameter object
    """

    name: str
    type: str
    contains: NotRequired[str]


class DomainDict(TypedDict):
    """
    TypedDict representing a domain object (both StarkNetDomain, StarknetDomain).
    """

    name: str
    version: str
    chainId: str
    revision: Optional[Revision]


class TypedDataDict(TypedDict):
    """
    TypedDict representing a TypedData object
    """

    types: Dict[str, List[ParameterDict]]
    primaryType: str
    domain: DomainDict
    message: Dict[str, Any]


class TypeContext(TypedDict):
    parent: str
    key: str



from abc import ABC, abstractmethod
from typing import List

from starknet_py.net.models.transaction import AccountTransaction
from starknet_py.utils.typed_data import TypedData


class BaseSigner(ABC):
    """
    Base class for transaction signer. Implement methods from this ABC to use a custom signer in Account.
    """

    @property
    @abstractmethod
    def public_key(self) -> int:
        """
        Public key of the signer.

        :return: public key
        """

    @abstractmethod
    def sign_transaction(self, transaction: AccountTransaction) -> List[int]:
        """
        Sign execute transaction and return a signature

        :param transaction: Execute transaction to sign
        :return: transaction signature
        """

    @abstractmethod
    def sign_message(self, typed_data: TypedData, account_address: int) -> List[int]:
        """
        Sign TypedData object for off-chain usage with the Starknet private key and return the signature.
        This adds a message prefix, so it can't be interchanged with transactions.

        :param typed_data: TypedData to be signed.
        :param account_address: account address.
        :return: the signature of the JSON object.
        """



from typing import cast

import pytest

from starknet_py.serialization.data_serializers import BoolSerializer
from starknet_py.serialization.errors import InvalidTypeException


@pytest.mark.parametrize(
    "value",
    [True, False],
)
def test_valid_bool_values(value):
    serialized = BoolSerializer().serialize(value)
    deserialized = BoolSerializer().deserialize([value])

    assert deserialized == value
    assert serialized == [value]


def test_invalid_type():
    error_message = "Error: expected bool, received '{}' of type '<class 'dict'>."
    with pytest.raises(InvalidTypeException, match=error_message):
        BoolSerializer().serialize(cast(bool, {}))



from typing import Tuple

import pytest
import pytest_asyncio

from starknet_py.cairo.felt import decode_shortstring
from starknet_py.contract import Contract, DeclareResult, DeployResult
from starknet_py.hash.storage import get_storage_var_address
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS
from starknet_py.tests.e2e.fixtures.misc import ContractVersion, load_contract

U128_MAX = (1 << 128) - 1
U256_MAX = (1 << 256) - 1


@pytest_asyncio.fixture(scope="package")
async def declare_deploy_hello2(account) -> Tuple[DeclareResult, DeployResult]:
    contract = load_contract(contract_name="Hello2", version=ContractVersion.V2)

    declare_result = await Contract.declare_v3(
        account=account,
        compiled_contract=contract["sierra"],
        compiled_contract_casm=contract["casm"],
        auto_estimate=True,
    )
    await declare_result.wait_for_acceptance()

    deploy_result = await declare_result.deploy_v3(auto_estimate=True)
    await deploy_result.wait_for_acceptance()

    return declare_result, deploy_result


@pytest_asyncio.fixture(scope="package", name="contract")
# pylint: disable=redefined-outer-name
async def hello2_contract(declare_deploy_hello2) -> Contract:
    _, deploy_result = declare_deploy_hello2
    return deploy_result.deployed_contract


@pytest.mark.asyncio
async def test_deploy_cairo2(contract):
    assert isinstance(contract, Contract)


@pytest.mark.asyncio
async def test_cairo2_interaction(contract):
    invoke_res = await contract.functions["increase_balance"].invoke_v3(
        amount=100, auto_estimate=True
    )
    await invoke_res.wait_for_acceptance()

    invoke_res = await contract.functions["increase_balance"].invoke_v3(
        amount=100, resource_bounds=MAX_RESOURCE_BOUNDS
    )
    await invoke_res.wait_for_acceptance()

    (balance,) = await contract.functions["get_balance"].call()
    assert balance == 200


@pytest.mark.asyncio
async def test_cairo2_interaction2(contract):
    invoke_res = await contract.functions["increase_balance_u8"].invoke_v3(
        255, auto_estimate=True
    )
    await invoke_res.wait_for_acceptance()

    (balance,) = await contract.functions["get_balance_u8"].call()
    assert balance == 255


@pytest.mark.asyncio
@pytest.mark.parametrize("uint_bits", [16, 32, 64, 128, 256])
async def test_cairo2_uint(contract, uint_bits):
    (result,) = await contract.functions[f"test_u{uint_bits}"].call(255)
    assert result == 256


@pytest.mark.asyncio
async def test_cairo2_u256(contract):
    (result,) = await contract.functions["test_u256"].call(p1=U256_MAX - 1)
    assert result == U256_MAX

    (result,) = await contract.functions["test_u256"].call(
        p1={"low": U128_MAX - 1, "high": U128_MAX}
    )
    assert result == U256_MAX


@pytest.mark.asyncio
async def test_cairo2_contract_address(contract):
    invoke_res = await contract.functions["set_ca"].invoke_v3(
        address=contract.account.address, auto_estimate=True
    )
    await invoke_res.wait_for_acceptance()

    (result,) = await contract.functions["get_ca"].call()
    assert result == contract.account.address


@pytest.mark.asyncio
async def test_cairo2_interaction3(contract):
    invoke_res = await contract.functions["increase_balance"].invoke_v3(
        100, auto_estimate=True
    )
    await invoke_res.wait_for_acceptance()
    (balance,) = await contract.functions["get_balance"].call()
    key = get_storage_var_address("balance")
    storage = await contract.client.get_storage_at(contract.address, key)
    assert storage == balance

    invoke_res = await contract.functions["set_ca"].invoke_v3(
        contract.account.address, auto_estimate=True
    )
    await invoke_res.wait_for_acceptance()
    (ca,) = await contract.functions["get_ca"].call()  # pylint: disable=invalid-name
    key = get_storage_var_address("ca")
    storage = await contract.client.get_storage_at(contract.address, key)
    assert storage == ca

    invoke_res = await contract.functions["set_status"].invoke_v3(
        True, auto_estimate=True
    )
    await invoke_res.wait_for_acceptance()
    (status,) = await contract.functions["get_status"].call()
    key = get_storage_var_address("status")
    storage = await contract.client.get_storage_at(contract.address, key)
    assert storage == status

    invoke_res = await contract.functions["set_user1"].invoke_v3(
        {
            "address": contract.account.address,
            "is_claimed": True,
        },
        auto_estimate=True,
    )
    await invoke_res.wait_for_acceptance()
    (user1,) = await contract.functions["get_user1"].call()
    key = get_storage_var_address("user1")
    storage1 = await contract.client.get_storage_at(contract.address, key)
    storage2 = await contract.client.get_storage_at(contract.address, key + 1)
    assert storage1 == user1["address"]
    assert storage2 == user1["is_claimed"]


@pytest.mark.asyncio
async def test_cairo2_echo(contract):
    (result,) = await contract.functions["echo_un_tuple"].call((77, 123))
    assert result == (77, 123)

    (result,) = await contract.functions["echo_array_u256"].call([123, 55, 77, 255])
    assert result == [123, 55, 77, 255]

    (result,) = await contract.functions["echo_array_bool"].call(
        [True, True, False, False]
    )
    assert result == [True, True, False, False]


@pytest.mark.asyncio
async def test_cairo2_echo_struct(contract):
    (result,) = await contract.functions["echo_struct"].call({"val": "simple"})
    assert decode_shortstring(result["val"]) == "simple"


@pytest.mark.asyncio
async def test_cairo2_echo_complex_struct(contract):
    invoke_result = await contract.functions["set_bet"].invoke_v3(auto_estimate=True)
    await invoke_result.wait_for_acceptance()

    (bet,) = await contract.functions["get_bet"].call(1)

    assert bet == {
        "name": 1952805748,
        "description": 6579555,
        "expire_date": 1,
        "creation_time": 1,
        "creator": contract.account.address,
        "is_cancelled": False,
        "is_voted": False,
        "bettor": {
            "address": contract.account.address,
            "is_claimed": False,
        },
        "counter_bettor": {
            "address": contract.account.address,
            "is_claimed": False,
        },
        "winner": False,
        "pool": 10,
        "amount": 1000,
    }


@pytest.mark.asyncio
async def test_cairo2_echo_tuple(contract):
    (result,) = await contract.functions["array_bool_tuple"].call([1, 2, 3], True)
    assert result == ([1, 2, 3, 1, 2], True)

    (result,) = await contract.functions["tuple_echo"].call(([1, 2, 3], [4, 5, 6]))
    assert result == ([1, 2, 3], [4, 5, 6])



import pytest

from starknet_py.serialization.data_serializers.array_serializer import ArraySerializer
from starknet_py.serialization.data_serializers.felt_serializer import FeltSerializer
from starknet_py.serialization.data_serializers.tuple_serializer import TupleSerializer

felt_array_serializer = ArraySerializer(FeltSerializer())


@pytest.mark.parametrize(
    "serializer, value, serialized_value",
    [
        (TupleSerializer([FeltSerializer(), FeltSerializer()]), (1, 2), [1, 2]),
        (
            TupleSerializer([FeltSerializer(), felt_array_serializer]),
            (1, [22, 38]),
            [1, 2, 22, 38],
        ),
        (
            # 3 nested tuples
            TupleSerializer([TupleSerializer([TupleSerializer([FeltSerializer()])])]),
            (((1,),),),
            [1],
        ),
        (
            TupleSerializer([FeltSerializer(), TupleSerializer([FeltSerializer()])]),
            (1, (2,)),
            [1, 2],
        ),
    ],
)
def test_valid_values(serializer, value, serialized_value):
    serialized = serializer.serialize(value)
    deserialized = serializer.deserialize(serialized_value)

    assert deserialized == value
    assert serialized == serialized_value



import pytest


@pytest.mark.asyncio
async def test_sign_offchain_message(account):
    # pylint: disable=import-outside-toplevel, duplicate-code, unused-variable

    # docs: start
    from starknet_py.net.account.account import Account
    from starknet_py.net.full_node_client import FullNodeClient
    from starknet_py.net.models import StarknetChainId
    from starknet_py.net.signer.key_pair import KeyPair
    from starknet_py.utils.typed_data import TypedData

    # Create a TypedData dictionary
    typed_data = {
        "types": {
            "StarkNetDomain": [
                {"name": "name", "type": "felt"},
                {"name": "version", "type": "felt"},
                {"name": "chainId", "type": "felt"},
            ],
            "Person": [
                {"name": "name", "type": "felt"},
                {"name": "wallet", "type": "felt"},
            ],
            "Mail": [
                {"name": "from", "type": "Person"},
                {"name": "to", "type": "Person"},
                {"name": "contents", "type": "felt"},
            ],
        },
        "primaryType": "Mail",
        "domain": {"name": "StarkNet Mail", "version": "1", "chainId": "1"},
        "message": {
            "from": {
                "name": "Cow",
                "wallet": "0xCD2a3d9F938E13CD947Ec05AbC7FE734Df8DD826",
            },
            "to": {
                "name": "Bob",
                "wallet": "0xbBbBBBBbbBBBbbbBbbBbbbbBBbBbbbbBbBbbBBbB",
            },
            "contents": "Hello, Bob!",
        },
    }
    # docs: end

    # save account fixture
    account_fixture = account

    # docs: start

    # Create an Account instance
    client = FullNodeClient(node_url="https://your.node.url")
    account = Account(
        client=client,
        address="0x1111",
        key_pair=KeyPair(private_key=123, public_key=456),
        chain=StarknetChainId.SEPOLIA,
    )
    # docs: end

    # retrieve account
    account = account_fixture

    # docs: start

    # Sign the message
    signature = account.sign_message(typed_data=typed_data)

    # Verify the message
    verify_result = account.verify_message(typed_data=typed_data, signature=signature)

    # Or if just a message hash is needed
    data = TypedData.from_dict(typed_data)
    message_hash = data.message_hash(account.address)

    # docs: end

    assert verify_result



from marshmallow import fields

from starknet_py.net.schemas.common import Felt
from starknet_py.utils.schema import Schema


class ContractsStorageKeysSchema(Schema):
    contract_address = Felt(data_key="contract_address", required=True)
    storage_keys = fields.List(Felt(), data_key="storage_keys", required=True)



from __future__ import annotations

from abc import ABC
from collections import OrderedDict
from dataclasses import dataclass
from typing import List


class CairoType(ABC):
    """
    Base type for all Cairo type representations. All types extend it.
    """


@dataclass
class FeltType(CairoType):
    """
    Type representation of Cairo field element.
    """


@dataclass
class BoolType(CairoType):
    """
    Type representation of Cairo boolean.
    """


@dataclass
class TupleType(CairoType):
    """
    Type representation of Cairo tuples without named fields.
    """

    types: List[CairoType]  #: Types of every tuple element.


@dataclass
class NamedTupleType(CairoType):
    """
    Type representation of Cairo tuples with named fields.
    """

    types: OrderedDict[str, CairoType]  #: Types of every tuple member.


@dataclass
class ArrayType(CairoType):
    """
    Type representation of Cairo arrays.
    """

    inner_type: CairoType  #: Type of element inside array.


@dataclass
class FixedSizeArrayType(CairoType):
    """
    Type representation of Cairo fixed-size arrays.
    """

    inner_type: CairoType  #: Type of element inside array.
    length: int  #: Length of the array.


@dataclass
class StructType(CairoType):
    """
    Type representation of Cairo structures.
    """

    name: str  #: Structure name
    # We need ordered dict, because it is important in serialization
    types: OrderedDict[str, CairoType]  #: types of every structure member.


@dataclass
class EnumType(CairoType):
    """
    Type representation of Cairo enums.
    """

    name: str  #: Enum name.
    variants: OrderedDict[str, CairoType]  #: Enum variants.


@dataclass
class OptionType(CairoType):
    """
    Type representation of Cairo options.
    """

    type: CairoType  #: Typed of element wrapped in the Option.


@dataclass
class IntType(CairoType):
    """
    Type representation of Cairo signed integers.
    """

    bits: int  #: Number of bits in the integer.


@dataclass
class UintType(CairoType):
    """
    Type representation of Cairo unsigned integers.
    """

    bits: int  #: Number of bits in the integer.

    def check_range(self, value: int):
        """
        Utility method checking if the `value` is in range.
        """


@dataclass
class TypeIdentifier(CairoType):
    """
    Type representation of Cairo identifiers.
    """

    name: str  #: Identifier name.


@dataclass
class UnitType(CairoType):
    """
    Type representation of Cairo unit `()`.
    """


@dataclass
class EventType(CairoType):
    """
    Type representation of Cairo Event.
    """

    name: str  #: Event name.
    types: OrderedDict[str, CairoType]  #: Types of every event member.
    keys: List[str]  #: Keys of every event member.


@dataclass
class NonZeroType(CairoType):
    """
    Type representation of Cairo NonZero.
    """

    type: CairoType  #: Type of element wrapped in NonZero.



# pylint: disable=import-outside-toplevel
import pytest

from starknet_py.constants import FIELD_PRIME


def test_generate_key_pair():
    # docs-generate: start
    from starknet_py.net.signer.key_pair import KeyPair

    key_pair = KeyPair.generate()
    # docs-generate: end

    hex_private_key = hex(key_pair.private_key)[2:]
    padded_hex_private_key = hex_private_key.zfill(64)
    assert len(padded_hex_private_key) == 64
    assert key_pair.private_key < FIELD_PRIME
    assert key_pair.public_key < FIELD_PRIME


def test_key_pair_from_private_key():
    # docs-from-private-key: start
    from starknet_py.net.signer.key_pair import KeyPair

    key_pair = KeyPair.from_private_key("0x1234abcd")
    # docs-from-private-key: end

    assert isinstance(key_pair.public_key, int)
    assert isinstance(key_pair.private_key, int)


@pytest.mark.skip(reason="This test requires a keystore file")
def test_key_pair_from_keystore():
    # docs-from-keystore: start
    from starknet_py.net.signer.key_pair import KeyPair

    key_pair = KeyPair.from_keystore("path/to/keystore", "password")
    # docs-from-keystore: end

    assert isinstance(key_pair.public_key, int)
    assert isinstance(key_pair.private_key, int)



import sys

import pytest

from starknet_py.contract import Contract
from starknet_py.hash.address import compute_address
from starknet_py.net.full_node_client import FullNodeClient
from starknet_py.net.udc_deployer.deployer import Deployer
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS
from starknet_py.utils.constructor_args_translator import translate_constructor_args


@pytest.mark.asyncio
async def test_default_deploy_with_class_hash(account, map_class_hash):
    deployer = Deployer()

    contract_deployment = deployer.create_contract_deployment(class_hash=map_class_hash)

    deploy_invoke_tx = await account.sign_invoke_v3(
        contract_deployment.call, resource_bounds=MAX_RESOURCE_BOUNDS
    )
    resp = await account.client.send_transaction(deploy_invoke_tx)
    await account.client.wait_for_tx(resp.transaction_hash)

    assert isinstance(contract_deployment.address, int)
    assert contract_deployment.address != 0


@pytest.mark.skipif(
    "--contract_dir=v2" not in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.asyncio
@pytest.mark.parametrize("calldata", [[10, 1, 2, 3, 3, 1, 2, 3, 12, 99]])
async def test_constructor_arguments_contract_deploy_without_abi(
    account,
    constructor_with_arguments_class_hash,
    calldata,
):
    deployer = Deployer(account_address=account.address)

    deploy_call, contract_address = deployer.create_contract_deployment(
        class_hash=constructor_with_arguments_class_hash,
        calldata=calldata,
    )

    deploy_invoke_transaction = await account.sign_invoke_v3(
        deploy_call, resource_bounds=MAX_RESOURCE_BOUNDS
    )
    resp = await account.client.send_transaction(deploy_invoke_transaction)
    await account.client.wait_for_tx(resp.transaction_hash)

    contract = await Contract.from_address(address=contract_address, provider=account)

    result = (await contract.functions["get"].call(block_number="latest"))[0]
    unwrapped_result = (result[0], result[1], result[2], dict(result[3]))
    expected_result = (
        10,
        (1, (2, 3)),
        sum([1, 2, 3]),
        {"value": 12, "nested_struct": {"value": 99}},
    )
    assert unwrapped_result == expected_result


@pytest.mark.skipif(
    "--contract_dir=v2" not in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "calldata",
    [
        [10, (1, (2, 3)), [1, 2, 3], {"value": 12, "nested_struct": {"value": 99}}],
        {
            "single_value": 10,
            "tuple": (1, (2, 3)),
            "arr": [1, 2, 3],
            "dict": {"value": 12, "nested_struct": {"value": 99}},
        },
    ],
)
async def test_constructor_arguments_contract_deploy(
    account,
    constructor_with_arguments_abi,
    constructor_with_arguments_class_hash,
    calldata,
):
    deployer = Deployer(account_address=account.address)

    deploy_call, contract_address = deployer.create_contract_deployment(
        class_hash=constructor_with_arguments_class_hash,
        abi=constructor_with_arguments_abi,
        calldata=calldata,
    )

    deploy_invoke_transaction = await account.sign_invoke_v3(
        deploy_call, resource_bounds=MAX_RESOURCE_BOUNDS
    )
    resp = await account.client.send_transaction(deploy_invoke_transaction)
    await account.client.wait_for_tx(resp.transaction_hash)

    contract = Contract(
        address=contract_address,
        abi=constructor_with_arguments_abi,
        provider=account,
    )

    result = (await contract.functions["get"].call(block_number="latest"))[0]
    unwarpped_result = (result[0], result[1], result[2], dict(result[3]))
    assert unwarpped_result == (
        10,
        (1, (2, 3)),
        sum([1, 2, 3]),
        {"value": 12, "nested_struct": {"value": 99}},
    )


@pytest.mark.skipif(
    "--contract_dir=v1" in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "calldata",
    [
        [10, (1, (2, 3)), [1, 2, 3], {"value": 12, "nested_struct": {"value": 99}}],
        {
            "single_value": 10,
            "tuple": (1, (2, 3)),
            "arr": [1, 2, 3],
            "dict": {"value": 12, "nested_struct": {"value": 99}},
        },
    ],
)
async def test_throws_when_calldata_provided_without_abi(
    account,
    constructor_with_arguments_class_hash,
    calldata,
):
    deployer = Deployer(account_address=account.address)

    with pytest.raises(
        ValueError,
        match="Argument calldata was provided without an ABI. It cannot be serialized.",
    ):
        deployer.create_contract_deployment(
            class_hash=constructor_with_arguments_class_hash, calldata=calldata
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "salt, pass_account_address", [(1, True), (2, False), (None, True), (None, False)]
)
async def test_address_computation(salt, pass_account_address, account, map_class_hash):
    deployer = Deployer(
        account_address=account.address if pass_account_address else None
    )
    if isinstance(salt, int) and isinstance(account.client, FullNodeClient):
        # transactions have to be different for each account
        salt += 1

    deploy_call, computed_address = deployer.create_contract_deployment(
        class_hash=map_class_hash,
        salt=salt,
    )

    deploy_invoke_tx = await account.sign_invoke_v3(
        deploy_call, resource_bounds=MAX_RESOURCE_BOUNDS
    )
    resp = await account.client.send_transaction(deploy_invoke_tx)
    await account.client.wait_for_tx(resp.transaction_hash)

    tx_receipt = await account.client.get_transaction_receipt(resp.transaction_hash)
    address_from_event = tx_receipt.events[0].data[0]

    assert computed_address == address_from_event


@pytest.mark.skipif(
    "--contract_dir=v2" not in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "calldata",
    [
        [10, (1, (2, 3)), [1, 2, 3], {"value": 12, "nested_struct": {"value": 99}}],
        {
            "single_value": 10,
            "tuple": (1, (2, 3)),
            "arr": [1, 2, 3],
            "dict": {"value": 12, "nested_struct": {"value": 99}},
        },
    ],
)
async def test_create_deployment_call_raw(
    account,
    constructor_with_arguments_abi,
    constructor_with_arguments_class_hash,
    calldata,
):
    deployer = Deployer(account_address=account.address)

    raw_calldata = translate_constructor_args(
        abi=constructor_with_arguments_abi, constructor_args=calldata
    )

    (
        deploy_call,
        contract_address,
    ) = deployer.create_contract_deployment_raw(
        class_hash=constructor_with_arguments_class_hash,
        raw_calldata=raw_calldata,
    )

    deploy_invoke_transaction = await account.sign_invoke_v3(
        deploy_call, resource_bounds=MAX_RESOURCE_BOUNDS
    )
    resp = await account.client.send_transaction(deploy_invoke_transaction)
    await account.client.wait_for_tx(resp.transaction_hash)

    assert isinstance(contract_address, int)
    assert contract_address != 0


@pytest.mark.skipif(
    "--contract_dir=v2" not in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.asyncio
async def test_create_deployment_call_raw_supports_seed_0(
    account,
    constructor_with_arguments_abi,
    constructor_with_arguments_class_hash,
):
    sample_calldata = {
        # the transactions have to be different for each account
        "single_value": 20 if isinstance(account.client, FullNodeClient) else 30,
        "tuple": (1, (2, 3)),
        "arr": [1, 2, 3],
        "dict": {"value": 12, "nested_struct": {"value": 99}},
    }
    deployer = Deployer()

    raw_calldata = translate_constructor_args(
        abi=constructor_with_arguments_abi,
        constructor_args=sample_calldata,
    )

    expected_address = compute_address(
        class_hash=constructor_with_arguments_class_hash,
        constructor_calldata=raw_calldata,
        salt=1,
    )

    (
        deploy_call,
        contract_address,
    ) = deployer.create_contract_deployment_raw(
        class_hash=constructor_with_arguments_class_hash,
        raw_calldata=raw_calldata,
        salt=1,
    )

    deploy_invoke_transaction = await account.sign_invoke_v3(
        deploy_call, resource_bounds=MAX_RESOURCE_BOUNDS
    )
    resp = await account.client.send_transaction(deploy_invoke_transaction)
    await account.client.wait_for_tx(resp.transaction_hash)

    assert isinstance(contract_address, int)
    assert (
        contract_address == expected_address
    ), f"Expected address {expected_address}, got {contract_address}."



import pytest

from starknet_py.serialization.data_serializers.unit_serializer import UnitSerializer

serializer = UnitSerializer()


def test_deserialize_unit():
    deserialized = serializer.deserialize([])

    assert deserialized is None


def test_serialize_unit():
    # pylint: disable=use-implicit-booleaness-not-comparison
    serialized = serializer.serialize(None)

    assert serialized == []


def test_throws_on_not_none():
    with pytest.raises(ValueError, match="Can only serialize `None`."):
        serializer.serialize("abc")  # type: ignore



import os

from marshmallow import EXCLUDE, RAISE
from marshmallow import Schema as MarshmallowSchema
from marshmallow import SchemaOpts

MARSHMALLOW_UNKNOWN_EXCLUDE = os.environ.get("STARKNET_PY_MARSHMALLOW_UNKNOWN_EXCLUDE")


class UnknownOpts(SchemaOpts):

    def __init__(self, meta, **kwargs):
        SchemaOpts.__init__(self, meta, **kwargs)
        self.unknown = (
            EXCLUDE if (MARSHMALLOW_UNKNOWN_EXCLUDE or "").lower() == "true" else RAISE
        )


# TODO(#1564): `ExcludeOpts` should be removed once issue is resolved.
class ExcludeOpts(SchemaOpts):

    def __init__(self, meta, **kwargs):
        SchemaOpts.__init__(self, meta, **kwargs)
        self.unknown = EXCLUDE


class Schema(MarshmallowSchema):
    OPTIONS_CLASS = UnknownOpts



# pylint: disable=import-outside-toplevel, no-member, duplicate-code


def test_synchronous_api(account, map_contract):
    # docs: start
    from starknet_py.contract import Contract
    from starknet_py.net.client_models import ResourceBounds, ResourceBoundsMapping

    contract_address = (
        "0x01336fa7c870a7403aced14dda865b75f29113230ed84e3a661f7af70fe83e7b"
    )
    # docs: end

    contract_address = map_contract.address

    # docs: start

    key = 1234
    contract = Contract.from_address_sync(address=contract_address, provider=account)

    invocation = contract.functions["put"].invoke_v3_sync(
        key,
        7,
        resource_bounds=ResourceBoundsMapping(
            l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
            l2_gas=ResourceBounds(max_amount=int(1e10), max_price_per_unit=int(1e17)),
            l1_data_gas=ResourceBounds(
                max_amount=int(1e5), max_price_per_unit=int(1e13)
            ),
        ),
    )
    invocation.wait_for_acceptance_sync()

    (saved,) = contract.functions["get"].call_sync(key)  # 7
    # docs: end

    assert saved == 7



import sys

import pytest

from starknet_py.tests.e2e.fixtures.misc import load_contract


@pytest.mark.skipif(
    "--contract_dir=v2" not in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.asyncio
async def test_simple_declare_and_deploy(account):
    # pylint: disable=import-outside-toplevel
    # docs: start
    from starknet_py.contract import Contract
    from starknet_py.net.client_models import ResourceBounds, ResourceBoundsMapping

    # docs: end
    compiled_contract = load_contract("AccountCopy1")
    constructor_args = {"public_key": 0x123}

    # docs: start
    declare_result = await Contract.declare_v3(
        account=account,
        compiled_contract=compiled_contract["sierra"],
        compiled_contract_casm=compiled_contract["casm"],
        resource_bounds=ResourceBoundsMapping(
            l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
            l2_gas=ResourceBounds(max_amount=int(1e9), max_price_per_unit=int(1e17)),
            l1_data_gas=ResourceBounds(
                max_amount=int(1e5), max_price_per_unit=int(1e13)
            ),
        ),
    )
    await declare_result.wait_for_acceptance()

    deploy_result = await declare_result.deploy_v3(
        constructor_args=constructor_args,
        resource_bounds=ResourceBoundsMapping(
            l1_gas=ResourceBounds(max_amount=int(1e5), max_price_per_unit=int(1e13)),
            l2_gas=ResourceBounds(max_amount=int(1e9), max_price_per_unit=int(1e17)),
            l1_data_gas=ResourceBounds(
                max_amount=int(1e5), max_price_per_unit=int(1e13)
            ),
        ),
    )
    await deploy_result.wait_for_acceptance()

    contract = deploy_result.deployed_contract
    # docs: end

    assert isinstance(declare_result.hash, int)
    assert isinstance(declare_result.class_hash, int)
    assert declare_result.compiled_contract == compiled_contract["sierra"]
    assert contract.address != 0



import json

from marshmallow import EXCLUDE

from starknet_py.abi.v0.schemas import ContractAbiEntrySchema
from starknet_py.tests.e2e.fixtures.constants import CAIRO_0_CONTRACTS_ABI_DIR
from starknet_py.tests.e2e.fixtures.misc import read_contract


def test_deserialize_abi():
    abi = json.loads(
        read_contract("complex_contract_abi.json", directory=CAIRO_0_CONTRACTS_ABI_DIR)
    )

    deserialized = [
        ContractAbiEntrySchema().load(entry, unknown=EXCLUDE) for entry in abi
    ]

    assert len(deserialized) == len(abi)



from collections import OrderedDict
from typing import Dict, Union

import pytest

from starknet_py.cairo.data_types import (
    ArrayType,
    EnumType,
    FeltType,
    StructType,
    TupleType,
    UnitType,
)
from starknet_py.cairo.v2.type_parser import TypeParser, UnknownCairoTypeError


@pytest.mark.parametrize(
    "type_string, expected",
    [
        ("core::felt252", FeltType()),
        ("core::array::Array::<core::felt252>", ArrayType(FeltType())),
        (
            "(core::felt252, core::felt252, core::felt252, core::felt252)",
            TupleType([FeltType()] * 4),
        ),
        ("(,)", TupleType([])),
        ("()", UnitType()),
        (
            "(core::felt252, (core::felt252, (core::array::Array::<core::felt252>, core::felt252)))",
            TupleType(
                [
                    FeltType(),
                    TupleType(
                        [
                            FeltType(),
                            TupleType(
                                [
                                    ArrayType(FeltType()),
                                    FeltType(),
                                ]
                            ),
                        ]
                    ),
                ]
            ),
        ),
    ],
)
def test_parse_without_defined_types(type_string, expected):
    parsed = TypeParser({}).parse_inline_type(type_string)
    assert parsed == expected


uint256_type = StructType("Uint256", OrderedDict(low=FeltType(), high=FeltType()))
wrapped_felt_type = StructType("WrappedFelt", OrderedDict(value=FeltType()))


@pytest.mark.parametrize(
    "type_string, expected",
    [
        ("Uint256", uint256_type),
        ("core::array::Array::<Uint256>", ArrayType(uint256_type)),
        (
            "(Uint256, WrappedFelt)",
            TupleType([uint256_type, wrapped_felt_type]),
        ),
        (
            "(Uint256, (WrappedFelt, (core::array::Array::<core::felt252>, core::array::Array::<WrappedFelt>)))",
            TupleType(
                [
                    uint256_type,
                    TupleType(
                        [
                            wrapped_felt_type,
                            TupleType(
                                [
                                    ArrayType(FeltType()),
                                    ArrayType(wrapped_felt_type),
                                ]
                            ),
                        ]
                    ),
                ]
            ),
        ),
    ],
)
def test_parse_with_defined_types(type_string, expected):
    types: Dict[str, Union[StructType, EnumType]] = {
        "Uint256": uint256_type,
        "WrappedFelt": wrapped_felt_type,
    }
    parsed = TypeParser(types).parse_inline_type(type_string)  # pyright: ignore
    assert parsed == expected


def test_missing_type():
    with pytest.raises(
        UnknownCairoTypeError, match="Type 'Uint256' is not defined"
    ) as err_info:
        TypeParser({}).parse_inline_type("Uint256")

    assert err_info.value.type_name == "Uint256"


def test_names_not_matching():
    with pytest.raises(
        ValueError, match="Keys must match name of type, 'OtherName' != 'Uint256'."
    ):
        TypeParser({"OtherName": uint256_type})



from typing import Optional, Type, Union

import pytest
from marshmallow import Schema, ValidationError

from starknet_py.net.client_models import BlockStatus, DAMode, Hash, TransactionStatus
from starknet_py.net.schemas.common import (
    BlockStatusField,
    DAModeField,
    Felt,
    NonPrefixedHex,
    StatusField,
    Uint64,
    Uint128,
)


class SchemaWithUint64(Schema):
    value = Uint64(data_key="value")


class SchemaWithUint128(Schema):
    value = Uint128(data_key="value")


class SchemaWithFelt(Schema):
    value = Felt(data_key="value")


class SchemaWithDAModeField(Schema):
    value = DAModeField(data_key="value")


def test_serialize_felt():
    data = {"value": 2137}

    serialized = SchemaWithFelt().dumps(data)
    assert '"value": "0x859"' in serialized


@pytest.mark.parametrize(
    "data",
    [
        {"value": None},
        {"value": 2**252},
    ],
)
def test_serialize_felt_throws_on_invalid_data(data):
    with pytest.raises(ValidationError, match="Invalid value provided for Felt"):
        SchemaWithFelt().dumps(data)


def test_deserialize_felt():
    data = {"value": "0x859"}

    deserialized = SchemaWithFelt().load(data)
    assert isinstance(deserialized, dict)
    assert deserialized["value"] == 2137


def test_deserialize_felt_throws_on_invalid_data():
    data = {"value": "2137"}

    with pytest.raises(ValidationError, match="Invalid value provided for Felt"):
        SchemaWithFelt().load(data)

    data = {"value": "0xwww"}
    with pytest.raises(ValidationError, match="Invalid value provided for Felt"):
        SchemaWithFelt().load(data)


@pytest.mark.parametrize(
    "data, expected_serialized",
    (
        ({"value": 0}, "0x0"),
        ({"value": "0x100"}, "0x100"),
        ({"value": 2**32}, "0x100000000"),
    ),
)
def test_serialize_uint64(data, expected_serialized):
    serialized = SchemaWithUint64().dumps(data)
    assert f'"value": "{expected_serialized}"' in serialized


@pytest.mark.parametrize(
    "data",
    [{"value": -1}, {"value": 2**64}, {"value": None}],
)
def test_serialize_uint64_throws_on_invalid_data(data):
    with pytest.raises(
        ValidationError,
        match=get_uint_error_message(Uint64, data["value"]),
    ):
        SchemaWithUint64().dumps(data)


@pytest.mark.parametrize(
    "data",
    [{"value": "0x100000000"}, {"value": 2**32}],
)
def test_deserialize_uint64(data):
    deserialized = SchemaWithUint64().load(data)
    assert isinstance(deserialized, dict)
    assert deserialized["value"] == 2**32


@pytest.mark.parametrize(
    "data",
    [
        {"value": -1},
        {"value": "1000"},
        {"value": 2**64},
        {"value": "0xwrong"},
        {"value": ""},
    ],
)
def test_deserialize_uint64_throws_on_invalid_data(data):
    with pytest.raises(
        ValidationError,
        match=get_uint_error_message(Uint64, data["value"]),
    ):
        SchemaWithUint64().load(data)


def test_serialize_uint128():
    data = {"value": 2**64}
    serialized = SchemaWithUint128().dumps(data)
    assert '"value": "0x10000000000000000"' in serialized


def test_serialize_uint128_throws_on_invalid_data():
    data = {"value": 2**128}
    with pytest.raises(
        ValidationError,
        match=get_uint_error_message(Uint128, data["value"]),
    ):
        SchemaWithUint128().dumps(data)


@pytest.mark.parametrize(
    "data",
    [{"value": "0x10000000000000000"}, {"value": 2**64}],
)
def test_deserialize_uint128(data):
    deserialized = SchemaWithUint128().load(data)
    assert isinstance(deserialized, dict)
    assert deserialized["value"] == 2**64


@pytest.mark.parametrize(
    "data",
    [
        {"value": -1},
        {"value": "1000"},
        {"value": 2**128},
        {"value": "0xwrong"},
        {"value": ""},
    ],
)
def test_deserialize_uint128_throws_on_invalid_data(data):
    with pytest.raises(
        ValidationError,
        match=get_uint_error_message(Uint128, data["value"]),
    ):
        SchemaWithUint128().load(data)


def test_serialize_hex():
    class SchemaWithHex(Schema):
        value1 = NonPrefixedHex(data_key="value1")

    data = {"value1": 123}

    serialized = SchemaWithHex().dump(data)
    assert isinstance(serialized, dict)
    assert serialized["value1"] == "7b"


def test_deserialize_hex():
    class SchemaWithHex(Schema):
        value1 = NonPrefixedHex(data_key="value1")

    data = {"value1": "7b"}

    deserialized = SchemaWithHex().load(data)
    assert isinstance(deserialized, dict)
    assert deserialized["value1"] == 123


def test_serialize_status_field():
    class SchemaWithStatusField(Schema):
        value1 = StatusField(data_key="value1")

    data = {"value1": TransactionStatus.RECEIVED}

    serialized = SchemaWithStatusField().dumps(data)
    assert '"value1": "RECEIVED"' in serialized


def test_deserialize_status_field():
    class SchemaWithStatusField(Schema):
        value1 = StatusField(data_key="value1")

    data = {"value1": "RECEIVED"}

    deserialized = SchemaWithStatusField().load(data)
    assert isinstance(deserialized, dict)
    assert deserialized["value1"] == TransactionStatus.RECEIVED


def test_deserialize_status_field_throws_on_invalid_data():
    class SchemaWithStatusField(Schema):
        value1 = StatusField(data_key="value1")

    data = {"value1": "SENT"}

    with pytest.raises(
        ValidationError, match="Invalid value provided for TransactionStatus"
    ):
        SchemaWithStatusField().load(data)


def test_serialize_block_status_field():
    class SchemaWithBlockStatusField(Schema):
        value1 = BlockStatusField(data_key="value1")

    data = {"value1": BlockStatus.PRE_CONFIRMED}

    serialized = SchemaWithBlockStatusField().dumps(data)
    assert '"value1": "PRE_CONFIRMED"' in serialized


def test_deserialize_block_status_field():
    class SchemaWithBlockStatusField(Schema):
        value1 = BlockStatusField(data_key="value1")

    data = {"value1": "PRE_CONFIRMED"}

    deserialized = SchemaWithBlockStatusField().load(data)
    assert isinstance(deserialized, dict)
    assert deserialized["value1"] == BlockStatus.PRE_CONFIRMED


def test_serialize_block_status_field_throws_on_invalid_data():
    class SchemaWithBlockStatusField(Schema):
        value1 = BlockStatusField(data_key="value1")

    data = {"value1": "SENT"}

    with pytest.raises(ValidationError, match="Invalid value for BlockStatus provided"):
        SchemaWithBlockStatusField().load(data)


@pytest.mark.parametrize(
    "data",
    [{"value": DAMode.L1}, {"value": DAMode.L2}],
)
def test_serialize_damode_field(data):
    serialized = SchemaWithDAModeField().dumps(data)
    assert f'"value": "{data["value"].name}"' in serialized


@pytest.mark.parametrize(
    "data, expected_deserialized",
    (
        ({"value": DAMode.L1.name}, DAMode.L1),
        ({"value": DAMode.L2.name}, DAMode.L2),
    ),
)
def test_deserialize_damode_field(data, expected_deserialized):
    deserialized = SchemaWithDAModeField().load(data)
    assert isinstance(deserialized, dict)
    assert deserialized["value"] == expected_deserialized


def get_uint_error_message(
    class_type: Union[Type[Uint64], Type[Uint128]], value: Optional[Hash]
) -> str:
    return f"Invalid value provided for {class_type.__name__}: {str(value)}"



# pylint: disable=redefined-outer-name

import json
import random
import sys
from enum import Enum
from pathlib import Path
from typing import Optional

import pytest

from starknet_py.net.client_models import (
    BlockStatus,
    DAMode,
    InvokeTransactionV3,
    L1DAMode,
    ResourceBounds,
    ResourceBoundsMapping,
    ResourcePrice,
    StarknetBlock,
)
from starknet_py.net.full_node_client import FullNodeClient
from starknet_py.net.models.typed_data import TypedDataDict
from starknet_py.tests.e2e.fixtures.constants import (
    CONTRACTS_V1_ARTIFACTS_MAP,
    CONTRACTS_V1_COMPILED,
    CONTRACTS_V2_ARTIFACTS_MAP,
    CONTRACTS_V2_COMPILED,
    TYPED_DATA_DIR,
)
from starknet_py.utils.typed_data import TypedData


def pytest_addoption(parser):
    parser.addoption(
        "--contract_dir",
        action="store",
        default="",
        help="Contract directory: possible 'v1', 'v2'",
    )


@pytest.fixture(
    params=[
        "typed_data_rev_0_example.json",
        "typed_data_rev_0_felt_array_example.json",
        "typed_data_rev_0_long_string_example.json",
        "typed_data_rev_0_struct_array_example.json",
        "typed_data_rev_1_example.json",
    ],
)
def typed_data(request) -> TypedDataDict:
    """
    Returns TypedData dictionary example.
    """
    file_name = getattr(request, "param")
    file_path = TYPED_DATA_DIR / file_name

    with open(file_path, "r", encoding="utf-8") as file:
        typed_data = json.load(file)

    return typed_data


@pytest.fixture()
def loaded_typed_data(request) -> TypedData:
    file_name = getattr(request, "param")
    file_path = TYPED_DATA_DIR / file_name

    with open(file_path, "r", encoding="utf-8") as file:
        typed_data_json = json.load(file)

    return TypedData.from_dict(typed_data_json)


@pytest.fixture(name="get_tx_receipt_path", scope="package")
def get_tx_receipt_full_node_client():
    return f"{FullNodeClient.__module__}.FullNodeClient.get_transaction_receipt"


@pytest.fixture(name="get_tx_status_path", scope="package")
def get_tx_status_full_node_client():
    return f"{FullNodeClient.__module__}.FullNodeClient.get_transaction_status"


@pytest.fixture(name="get_block_with_txs_path", scope="package")
def get_block_with_txs_full_node_client():
    return f"{FullNodeClient.__module__}.FullNodeClient.get_block_with_txs"


def transaction_mock_with_tip(tip: int) -> InvokeTransactionV3:
    return InvokeTransactionV3(
        hash=random.randint(0, 1000),
        signature=[],
        version=3,
        resource_bounds=ResourceBoundsMapping(
            l1_gas=ResourceBounds(max_amount=1, max_price_per_unit=1),
            l2_gas=ResourceBounds(max_amount=1, max_price_per_unit=1),
            l1_data_gas=ResourceBounds(max_amount=1, max_price_per_unit=1),
        ),
        paymaster_data=[],
        nonce_data_availability_mode=DAMode.L2,
        fee_data_availability_mode=DAMode.L2,
        calldata=[],
        sender_address=random.randint(0, 100000),
        nonce=1,
        account_deployment_data=[],
        tip=tip,
    )


def starknet_block_mock() -> StarknetBlock:
    return StarknetBlock(
        block_number=1,
        block_hash=0x123,
        parent_hash=0x111,
        new_root=0x222,
        timestamp=1640000000,
        sequencer_address=0x1,
        l1_gas_price=ResourcePrice(price_in_fri=1, price_in_wei=1),
        l2_gas_price=ResourcePrice(price_in_fri=1, price_in_wei=1),
        l1_data_gas_price=ResourcePrice(price_in_fri=1, price_in_wei=1),
        l1_da_mode=L1DAMode.BLOB,
        starknet_version="0.14.0",
        status=BlockStatus.ACCEPTED_ON_L2,
        transactions=[],
    )


@pytest.fixture(name="block_with_tips_mock", scope="package")
def block_with_tips_mock():
    block = starknet_block_mock()
    block.transactions = [
        transaction_mock_with_tip(2),
        transaction_mock_with_tip(1),
        transaction_mock_with_tip(3),
    ]
    return block


class ContractVersion(Enum):
    V1 = "V1"
    V2 = "V2"


class UnknownArtifacts(BaseException):
    pass


def load_contract(contract_name: str, version: Optional[ContractVersion] = None):
    if version is None:
        if "--contract_dir=v1" in sys.argv:
            version = ContractVersion.V1
        if "--contract_dir=v2" in sys.argv:
            version = ContractVersion.V2

    if version == ContractVersion.V1:
        artifacts_map_path = CONTRACTS_V1_ARTIFACTS_MAP
        directory = CONTRACTS_V1_COMPILED
    else:
        artifacts_map_path = CONTRACTS_V2_ARTIFACTS_MAP
        directory = CONTRACTS_V2_COMPILED

    artifacts_map = json.loads((artifacts_map_path).read_text("utf-8"))

    artifact_file_names = next(
        (
            item["artifacts"]
            for item in artifacts_map["contracts"]
            if item["contract_name"] == contract_name
        ),
        None,
    )

    if not isinstance(artifact_file_names, dict):  # pyright: ignore
        raise UnknownArtifacts(f"Artifacts for contract {contract_name} not found")

    sierra = (directory / artifact_file_names["sierra"]).read_text("utf-8")
    casm = (directory / artifact_file_names["casm"]).read_text("utf-8")

    return {"casm": casm, "sierra": sierra}


def read_contract(file_name: str, *, directory: Path) -> str:
    """
    Return contents of file_name from directory.
    """
    return (directory / file_name).read_text("utf-8")



# pylint: disable=unused-variable
from starknet_py.net.udc_deployer.deployer import Deployer


def test_init():
    # docs-start: init
    deployer = Deployer()
    # or
    deployer = Deployer(deployer_address=0x123)
    # or
    deployer = Deployer(deployer_address=0x123, account_address=0x321)
    # docs-end: init


def test_create_contract_deployment_raw():
    deployer = Deployer()

    # docs-start: create_contract_deployment_raw
    contract_deployment = deployer.create_contract_deployment_raw(
        class_hash=0x123, salt=1, raw_calldata=[3, 1, 2, 3]
    )
    # docs-end: create_contract_deployment_raw
